# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides shared logic for publishing diagnostics for type errors
"""

import dataclasses
import logging
from pathlib import Path
from typing import Collection, Dict, List, Optional

from .. import backend_arguments, error, json_rpc
from ..language_server import connections, protocol as lsp
from . import error_code_to_link_mapper, server_state

PYRE_DOCUMENTATION_LINK = "https://pyre-check.org/docs/errors/#"
LOG: logging.Logger = logging.getLogger(__name__)


async def _publish_diagnostics(
    output_channel: connections.AsyncTextWriter,
    path: Path,
    diagnostics: Collection[lsp.Diagnostic],
) -> None:
    LOG.debug(f"Publish diagnostics for {path}: {diagnostics}")
    await lsp.write_json_rpc(
        output_channel,
        json_rpc.Request(
            method="textDocument/publishDiagnostics",
            parameters=json_rpc.ByNameParameters(
                {
                    "uri": lsp.DocumentUri.from_file_path(path).unparse(),
                    "diagnostics": [diagnostic.to_dict() for diagnostic in diagnostics],
                }
            ),
        ),
    )


def type_errors_to_diagnostics(
    type_errors: Collection[error.Error],
    set_unused_as_warning: bool,
) -> Dict[Path, List[lsp.Diagnostic]]:
    result: Dict[Path, List[lsp.Diagnostic]] = {}
    for type_error in type_errors:
        result.setdefault(type_error.path, []).append(
            type_error_to_diagnostic(type_error, set_unused_as_warning)
        )
    return result


def type_error_to_diagnostic(
    type_error: error.Error, set_unused_as_warning: bool
) -> lsp.Diagnostic:
    code_description = _get_code_description(type_error)
    severity = (
        lsp.DiagnosticSeverity.WARNING
        if set_unused_as_warning and type_error.code == 0
        else lsp.DiagnosticSeverity.ERROR
    )
    return lsp.Diagnostic(
        range=lsp.LspRange(
            start=lsp.LspPosition(
                line=type_error.line - 1, character=type_error.column
            ),
            end=lsp.LspPosition(
                line=type_error.stop_line - 1, character=type_error.stop_column
            ),
        ),
        message=type_error.description,
        severity=severity,
        code=None if code_description is None else "pyre (documentation link)",
        code_description=code_description,
        source="Pyre",
    )


def _get_code_description(type_error: error.Error) -> lsp.CodeDescription:
    code = type_error.code
    fragment = error_code_to_link_mapper.error_code_to_fragment.get(code, "")
    href = PYRE_DOCUMENTATION_LINK + fragment
    return lsp.CodeDescription(href=href)


@dataclasses.dataclass(frozen=True)
class ClientTypeErrorHandler:
    client_output_channel: connections.AsyncTextWriter
    server_state: server_state.ServerState
    remote_logging: Optional[backend_arguments.RemoteLogging] = None

    def update_type_errors(
        self, type_errors: Collection[error.Error], set_unused_as_warning: bool = False
    ) -> None:
        LOG.info(
            "Refreshing type errors received from Pyre server. "
            f"Total number of type errors is {len(type_errors)}."
        )
        self.server_state.diagnostics = type_errors_to_diagnostics(
            type_errors, set_unused_as_warning
        )

    async def clear_type_errors_for_client(self) -> None:
        for path in self.server_state.diagnostics:
            await _publish_diagnostics(self.client_output_channel, path, [])

    async def show_type_errors_to_client(self) -> None:
        for path, diagnostics in self.server_state.diagnostics.items():
            await _publish_diagnostics(self.client_output_channel, path, diagnostics)

    async def show_overlay_type_errors(
        self,
        path: Path,
        type_errors: Collection[error.Error],
        set_unused_as_warning: bool = False,
    ) -> None:
        LOG.info(
            f"Refreshing type errors at path {path}. "
            f"Total number of type errors is {len(type_errors)}."
        )
        diagnostics_by_path = type_errors_to_diagnostics(
            type_errors, set_unused_as_warning
        )
        diagnostics = diagnostics_by_path.get(path, [])
        await _publish_diagnostics(self.client_output_channel, path, diagnostics)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module contains logic for `pyre init`, which interactively
helps a user create a new pyre configuration.
"""

import json
import logging
import os
import shutil
import subprocess
import sys
from logging import Logger
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

from .. import log
from ..find_directories import (
    BINARY_NAME,
    find_global_root,
    find_parent_directory_containing_file,
    find_taint_models_directories,
    find_typeshed,
    JSON_CONFIGURATION_FILE,
    LOCAL_CONFIGURATION_FILE,
)
from . import commands

LOG: Logger = logging.getLogger(__name__)


class InitializationException(Exception):
    pass


def _create_source_directory_element(source: str) -> Union[str, Dict[str, str]]:
    if source == ".":
        return source
    if not Path(source).is_dir():
        raise InitializationException(f"No directory found at `{source}`.")
    # Imports are likely relative to the parent of the package.
    package_root = find_parent_directory_containing_file(Path(source), "__init__.py")
    if package_root is not None:
        return {
            "import_root": os.path.relpath(str(package_root.parent), "."),
            "source": source,
        }
    else:
        return source


def _check_configuration_file_location(
    configuration_path: Path, current_directory: Path, global_root: Optional[Path]
) -> None:
    if os.path.isfile(configuration_path):
        if global_root:
            error = (
                "Local configurations must be created in subdirectories of "
                + f"`{str(current_directory)}` as it already contains a "
                + "`.pyre_configuration`."
            )
        else:
            error = (
                "A pyre configuration already exists at "
                + f"`{str(configuration_path)}`."
            )
        raise InitializationException(error)
    local_configuration_path = current_directory / LOCAL_CONFIGURATION_FILE
    if local_configuration_path.is_file():
        raise InitializationException(
            "A local pyre configuration already exists at "
            + f"`{str(local_configuration_path)}`."
        )


def _get_local_configuration(
    current_directory: Path, buck_root: Optional[Path]
) -> Dict[str, Any]:
    configuration: Dict[str, Any] = {}
    using_targets = log.get_yes_no_input("Is your project built with Buck?")
    if using_targets:
        targets = log.get_input(
            "Which buck target(s) should pyre analyze?\n"
            + "  Default: Analyze all targets under the configuration (assume fbcode).\n"
            + "  (Ex. `fbcode//target:a, fbsource//target/b/...`)\n"
        ).strip()
        if len(targets) == 0:
            if buck_root:
                root = current_directory.relative_to(buck_root)
                configuration["targets"] = [f"fbcode//{str(root)}/..."]
            else:
                raise InitializationException(
                    "No `.buckconfig` found with which to create a default target."
                )
        else:
            configuration["targets"] = [target.strip() for target in targets.split(",")]
    else:
        source_directories = log.get_input(
            "Which directory(ies) should pyre analyze?\n"
        )
        configuration["source_directories"] = [
            directory.strip() for directory in source_directories.split(",")
        ]

    # TODO(T132432706): Ask for oncall in global configuration, but not in OSS.
    oncall = log.get_input("What oncall is responsible for this project?\n").strip()
    if oncall:
        configuration["oncall"] = oncall
    return configuration


def _create_watchman_configuration() -> None:
    watchman_configuration_path = os.path.abspath(".watchmanconfig")
    watchman_path = shutil.which("watchman")
    if watchman_path is not None and log.get_yes_no_input(
        "Also initialize watchman in the current directory?"
    ):
        try:
            if not os.path.isfile(watchman_configuration_path):
                with open(watchman_configuration_path, "w+") as configuration_file:
                    configuration_file.write("{}\n")
                LOG.warning(
                    "Created basic `.watchmanconfig` at "
                    + f"{watchman_configuration_path}"
                )
            subprocess.run(
                [watchman_path, "watch-project", "."],
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            LOG.warning("Current directory is being watched by `watchman`.")
        except IsADirectoryError:
            LOG.warning(f"Unable to write to {watchman_configuration_path}.")
        except subprocess.CalledProcessError:
            LOG.warning("Failed to run `watchman watch-project .`.")


def _get_configuration(
    taint_models_directory_required: Optional[bool] = False,
) -> Dict[str, Any]:
    configuration: Dict[str, Any] = {}

    _create_watchman_configuration()
    binary_path = shutil.which(BINARY_NAME)
    if binary_path is None:
        binary_path = shutil.which(
            os.path.join(os.path.dirname(sys.argv[0]), BINARY_NAME)
        )
    if binary_path is None:
        binary_path = os.path.abspath(
            log.get_input(f"No `{BINARY_NAME}` found, enter the path manually: ")
        )
        if not os.path.isfile(binary_path):
            raise InitializationException(
                f"Unable to locate binary at `{binary_path}`."
            )
        configuration["binary"] = binary_path
    else:
        LOG.info(f"Binary found at `{binary_path}`")

    typeshed: Optional[Path] = find_typeshed()
    if typeshed is None:
        typeshed = Path(
            log.get_input("Unable to locate typeshed, please enter its root: ")
        ).resolve()
        if not typeshed.is_dir():
            raise InitializationException(
                f"No typeshed directory found at `{typeshed}`."
            )
        configuration["typeshed"] = str(typeshed)
    else:
        LOG.info(f"Typeshed found at `{typeshed}``")

    if taint_models_directory_required:
        taint_models_paths = find_taint_models_directories()
        if taint_models_paths is None:
            taint_models_paths = [
                Path(
                    log.get_input(
                        "Unable to find taint models directory, please enter its root: "
                    )
                ).resolve()
            ]
        configuration["taint_models_path"] = [str(path) for path in taint_models_paths]

    source_directory_input = log.get_optional_input(
        "Which directory(ies) should pyre analyze?", "."
    )
    source_directory_paths = [
        directory.strip() for directory in source_directory_input.split(",")
    ]
    configuration["source_directories"] = [
        _create_source_directory_element(path) for path in source_directory_paths
    ]

    LOG.info(
        "Pyre will automatically add typed pacakages installed on your system "
        "as type checking dependencies, according to PEP561. You can edit the "
        "configuration file if you want to change the behavior. "
    )
    configuration["site_package_search_strategy"] = "pep561"
    return configuration


def get_configuration_and_path(
    taint_models_directory_required: Optional[bool] = False,
) -> Tuple[Dict[str, Any], Path]:
    global_root: Optional[Path] = find_global_root(Path("."))
    buck_root: Optional[Path] = find_parent_directory_containing_file(
        Path("."), ".buckconfig"
    )
    current_directory: Path = Path(os.getcwd())
    configuration_path = current_directory / JSON_CONFIGURATION_FILE
    _check_configuration_file_location(
        configuration_path, current_directory, global_root
    )
    local_configuration_path = current_directory / LOCAL_CONFIGURATION_FILE
    if global_root:
        configuration_path = local_configuration_path
        configuration = _get_local_configuration(current_directory, buck_root)
    else:
        configuration = _get_configuration(taint_models_directory_required)
    return configuration, configuration_path


def write_configuration(
    configuration: Dict[str, Any], configuration_path: Path
) -> None:
    with open(configuration_path, "w+") as configuration_file:
        json.dump(configuration, configuration_file, sort_keys=True, indent=2)
        configuration_file.write("\n")


def run() -> commands.ExitCode:
    try:
        configuration, configuration_path = get_configuration_and_path()
        write_configuration(configuration, configuration_path)
        LOG.log(
            log.SUCCESS,
            "Successfully initialized pyre!\n"
            + f"  You can view the configuration at `{configuration_path}`.\n"
            + "  You can now run the type checker with `pyre`.",
        )
        return commands.ExitCode.SUCCESS
    except InitializationException as error:
        LOG.error(f"{error}")
        return commands.ExitCode.FAILURE



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
The daemon connection module contains low-level APIs for sending requests to the Pyre daemon.
This module is responsible for the communication with the Pyre daemon, which is
at the layer as low as the direct socket connection. The socket connection is used
to send a string form request to the daemon and returns the response in string form.

This module is NOT responsible for any kind of processing that happens to the request/response,
which is left to the higher layers.
"""

from __future__ import annotations

import asyncio
import contextlib
import dataclasses
import logging
import traceback
from pathlib import Path
from typing import AsyncIterator, Optional, Union

from .. import dataclasses_json_extensions as json_mixins, log
from . import connections

LOG: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class DaemonConnectionFailure(json_mixins.CamlCaseAndExcludeJsonMixin):
    error_message: str
    error_source: Optional[Exception] = None


def send_raw_request(socket_path: Path, raw_request: str) -> str:
    with connections.connect(socket_path) as (
        input_channel,
        output_channel,
    ):
        LOG.debug(f"Sending `{log.truncate(raw_request, 400)}`")
        output_channel.write(f"{raw_request}\n")
        raw_response = input_channel.readline().strip()
        LOG.debug(f"Received `{log.truncate(raw_response, 400)}`")
        return raw_response


@dataclasses.dataclass(frozen=True)
class AsyncConnection:
    reader: connections.AsyncTextReader
    writer: connections.AsyncTextWriter

    async def _send_request_no_logging(
        self,
        request: str,
    ) -> str:
        await self.writer.write(f"{request}\n")
        return await self.reader.read_until(separator="\n")

    async def send_request(
        self,
        request: str,
    ) -> str:
        LOG.debug(f"Sending `{log.truncate(request, 400)}`")
        response = await self._send_request_no_logging(request)
        LOG.info(f"Received: `{log.truncate(response, 400).strip()}`")
        return response

    @staticmethod
    @contextlib.asynccontextmanager
    async def create(
        socket_path: Path,
    ) -> AsyncIterator[AsyncConnection]:
        async with connections.connect_async(socket_path) as (reader, writer):
            yield AsyncConnection(reader, writer)


async def send_async_raw_request(
    socket_path: Path,
    request: str,
) -> str:
    async with AsyncConnection.create(socket_path) as connection:
        return await connection.send_request(request)


async def attempt_send_async_raw_request(
    socket_path: Path,
    request: str,
) -> Union[DaemonConnectionFailure, str]:
    try:
        return await send_async_raw_request(
            socket_path,
            request,
        )
    except (
        connections.ConnectionFailure,
        asyncio.IncompleteReadError,
        ConnectionError,
    ) as error:
        return DaemonConnectionFailure(
            error_message="Could not establish connection with an existing Pyre server "
            f"at {socket_path}: {error}. Type: {type(error)}. Stacktrace: {traceback.format_exc( limit = None, chain = True)}",
            error_source=error,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides a decorator called ``dataclass_merge``. It is intended to decorate
any dataclasses:

  @dataclass_merge
  @dataclass(frozen=True)
  class Foo:
      x: int | None = None
      y: int | None = None

The decorator will introduce an additional ``merge`` static method to the decorated
class:

  classs Foo:
      @staticmethod
      def merge(base: Foo, override: Foo) -> Foo:
          ...

The ``merge`` function attempts to create a "combined" object from its two arguments.
The combination procedure works by looking at each individual field of the two input
dataclass instances, merge the field of the two instances, and assemble the returned
object from all merged fields.

By default, field merging follow the "overwrite" policy, where fields in the
``override`` argument takes priority if it is not ``None``. For example:

  Foo.merge(base=Foo(x=1, y=2), override=Foo(x=3, y=4))     # Returns Foo(x=3, y=4)
  Foo.merge(base=Foo(x=1, y=2), override=Foo(x=3, y=None))  # Returns Foo(x=3, y=2)
  Foo.merge(base=Foo(x=1, y=2), override=Foo(x=None, y=4))  # Returns Foo(x=1, y=4)
  Foo.merge(base=Foo(x=1, y=2), override=Foo())             # Returns Foo(x=1, y=2)

Nested dataclasses will be merged recursively by default:

  @dataclass_merge
  @dataclass(frozen=True)
  class Bar:
      a: Foo
      b: str | None = None

  Bar.merge(Bar(Foo(x=1, y=2), "bar"), Bar(Foo(x=3, y=None)))
  # Returns Bar(a=Foo(x=3, y=2), b="bar")

To fine-tune merging policies, one can attach the ``merge_policy`` metadata to the
corresponding dataclass field:

  from dataclasses import dataclass, field
  from dataclasses_merge import dataclass_merge, Policy

  @dataclass_merge
  @dataclass(frozen=True)
  class Baz:
      x: int | None = field(
          default=None,
          metadata={"merge_policy": Policy.OVERWRITE}
      )
      y: List[int] = field(
          default_factory=list,
          metadata={"merge_policy": Policy.PREPEND}
      )
      z: str | None = field(
          default=None,
          metadata={"merge_policy": Policy.RAISE_WHEN_OVERWRITTEN}
      )
      w: Foo = field(
          default_factory=Foo,
          metadata={"merge_policy": Foo.merge}
      )

The "prepend" policy only works on list-typed fields. It prepends the list in
``override`` to the list in ``base``:

  Baz.merge(Baz(y=[1,2]), Baz(y=[3,4]))  # Returns Baz(y=[3,4,1,2])

The "raise_when_overwritten" policy forbids any overwriting: either the field
in ``base`` or the field in ``override`` can be ``None``, but not both:

  Baz.merge(Baz(z=1), Baz(z=None))     # OK
  Baz.merge(Baz(z=None), Baz(z=1))     # OK
  Baz.merge(Baz(z=None), Baz(z=None))  # OK
  Baz.merge(Baz(z=1), Baz(z=2))        # Excpetion raised!

And finally, clients can provide their own custom merge policy, by passing in a
callable to the ``merge_policy`` metadata. The provided callable will be served
as the "merge" function of the corresponding field.

All APIs provided in this module will only ever attempt to raise the
``DataclassMergeError`` exception. Callbacks passed as ``merge_policy`` are not
subject to this constraint, though.
"""

import dataclasses
import enum
from typing import Iterable, Optional, Type, TypeVar

T = TypeVar("T")


class Policy(str, enum.Enum):
    _value_: str
    OVERWRITE = "overwrite"
    PREPEND = "prepend"
    RAISE_WHEN_OVERWRITTEN = "raise_when_overwritten"


class DataclassMergeError(Exception):
    pass


def _is_dataclass_instance(instance: object) -> bool:
    return dataclasses.is_dataclass(instance) and not isinstance(instance, type)


def _assert_is_dataclass_instance(instance: object) -> None:
    if not _is_dataclass_instance(instance):
        raise DataclassMergeError(
            f"Expected dataclass instace but got `{type(instance).__name__}`"
        )


def _get_field(name: str, instance: object) -> object:
    if not hasattr(instance, name):
        raise DataclassMergeError(
            f"Field `{name}` does not exist on object of type {type(instance).__name__}"
        )
    return getattr(instance, name)


def _merge_fields(
    fields: "Iterable[dataclasses.Field[object]]",
    base: object,
    override: object,
) -> Iterable[object]:
    def overwrite(base: Optional[T], override: Optional[T]) -> Optional[T]:
        return base if override is None else override

    def prepend(base: object, override: object, name: str) -> object:
        if not isinstance(base, list):
            raise DataclassMergeError(
                f"Field `{name}` is expected to be a list as it uses the 'prepend' "
                f"merge policy. Got `{base}`."
            )
        if not isinstance(override, list):
            raise DataclassMergeError(
                f"Field `{name}` is expected to be a list as it uses the 'prepend' "
                f"merge policy. Got `{override}`."
            )
        return override + base

    def raise_when_overwritten(
        base: Optional[T], override: Optional[T], name: str
    ) -> Optional[T]:
        if base is None:
            return override
        elif override is None:
            return base
        else:
            raise DataclassMergeError(
                f"Field `{name}` is not allowed to be overwritten."
            )

    def default_policy(cls: Type[object], base: object, override: object) -> object:
        if hasattr(cls, "merge"):
            # pyre-ignore[16]: Pyre does not recognize hasattr() check
            return cls.merge(base, override)
        return overwrite(base, override)

    def merge(field: "dataclasses.Field[object]") -> object:
        field_name = field.name
        base_value = _get_field(field_name, base)
        override_value = _get_field(field_name, override)

        merge_policy = field.metadata.get("merge_policy", None)
        if merge_policy == Policy.RAISE_WHEN_OVERWRITTEN:
            return raise_when_overwritten(base_value, override_value, field_name)
        elif merge_policy == Policy.PREPEND:
            return prepend(base_value, override_value, field_name)
        elif merge_policy == Policy.OVERWRITE:
            return overwrite(base_value, override_value)
        elif merge_policy is not None:
            return merge_policy(base_value, override_value)
        else:
            return default_policy(field.type, base_value, override_value)

    return (merge(field) for field in fields)


def dataclass_merge(cls: Type[T]) -> Type[T]:
    def merge(base: T, override: T) -> T:
        _assert_is_dataclass_instance(base)
        _assert_is_dataclass_instance(override)
        return cls(
            *_merge_fields(
                dataclasses.fields(cls),
                base,
                override,
            )
        )

    cls.merge = staticmethod(merge)
    return cls



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module defines shared logic for data sent to the Pyre backend
as a json arguments file when launching a new backend process.

Several commands define their own specific arguments types with extra
fields, but most of these contain a `base_arguments` field of type
backend_arguments.BaseArguments.
"""

import contextlib
import dataclasses
import json
import logging
import os
import shutil
import tempfile
from pathlib import Path
from typing import Any, Dict, IO, Iterator, List, Optional, Sequence, Set, Union

from typing_extensions import Protocol

from . import (
    configuration as configuration_module,
    find_directories,
    frontend_configuration,
    identifiers,
)
from .configuration import search_path

LOG: logging.Logger = logging.getLogger(__name__)

SERVER_ARTIFACT_ROOT_NAME: str = "link_trees"


@dataclasses.dataclass(frozen=True)
class RemoteLogging:
    logger: str
    identifier: str = ""

    @staticmethod
    def create(
        logger: Optional[str] = None, identifier: Optional[str] = None
    ) -> "Optional[RemoteLogging]":
        return (
            RemoteLogging(logger=logger, identifier=identifier or "")
            if logger is not None
            else None
        )

    def serialize(self) -> Dict[str, str]:
        return {"logger": self.logger, "identifier": self.identifier}


@dataclasses.dataclass(frozen=True)
class SimpleSourcePath:
    elements: Sequence[search_path.Element] = dataclasses.field(default_factory=list)

    def serialize(self) -> Dict[str, object]:
        return {
            "kind": "simple",
            "paths": [element.command_line_argument() for element in self.elements],
        }

    def get_checked_directory_allowlist(self) -> Set[str]:
        return {element.path() for element in self.elements}

    def cleanup(self) -> None:
        pass


@dataclasses.dataclass(frozen=True)
class WithUnwatchedDependencySourcePath:
    change_indicator_root: Path
    unwatched_dependency: configuration_module.UnwatchedDependency
    elements: Sequence[search_path.Element] = dataclasses.field(default_factory=list)

    def serialize(self) -> Dict[str, object]:
        return {
            "kind": "with_unwatched_dependency",
            "paths": [element.command_line_argument() for element in self.elements],
            "unwatched_dependency": {
                "change_indicator": {
                    "root": str(self.change_indicator_root),
                    "relative": self.unwatched_dependency.change_indicator,
                },
                "files": {
                    "root": self.unwatched_dependency.files.root,
                    "checksum_path": self.unwatched_dependency.files.checksum_path,
                },
            },
        }

    def get_checked_directory_allowlist(self) -> Set[str]:
        return {element.path() for element in self.elements}

    def cleanup(self) -> None:
        pass


@dataclasses.dataclass(frozen=True)
class BuckSourcePath:
    source_root: Path
    artifact_root: Path
    checked_directory: Path
    targets: Sequence[str] = dataclasses.field(default_factory=list)
    targets_fallback_sources: Optional[Sequence[search_path.Element]] = None
    mode: Optional[str] = None
    isolation_prefix: Optional[str] = None
    bxl_builder: Optional[str] = None
    use_buck2: bool = True
    kill_buck_after_build: bool = False
    number_of_threads: Optional[int] = None

    def serialize(self) -> Dict[str, object]:
        mode = self.mode
        isolation_prefix = self.isolation_prefix
        bxl_builder = self.bxl_builder
        targets_fallback_sources = self.targets_fallback_sources
        return {
            "kind": "buck",
            "targets": self.targets,
            **(
                {}
                if targets_fallback_sources is None
                else {
                    "targets_fallback_sources": [
                        element.command_line_argument()
                        for element in targets_fallback_sources
                    ],
                }
            ),
            **({} if mode is None else {"mode": mode}),
            **(
                {}
                if isolation_prefix is None
                else {"isolation_prefix": isolation_prefix}
            ),
            **({} if bxl_builder is None else {"bxl_builder": bxl_builder}),
            "use_buck2": self.use_buck2,
            "source_root": str(self.source_root),
            "artifact_root": str(self.artifact_root),
            "kill_buck_after_build": self.kill_buck_after_build,
            **(
                {}
                if self.number_of_threads is None
                else {"number_of_threads": self.number_of_threads}
            ),
        }

    def get_checked_directory_allowlist(self) -> Set[str]:
        return {str(self.checked_directory)}

    def cleanup(self) -> None:
        shutil.rmtree(str(self.artifact_root), ignore_errors=True)


SourcePath = Union[SimpleSourcePath, WithUnwatchedDependencySourcePath, BuckSourcePath]


@dataclasses.dataclass(frozen=True)
class BaseArguments:
    """
    Data structure for configuration options common to many backend commands.
    Need to keep in sync with `pyre/source/command/commandStartup.ml`
    """

    log_path: str
    global_root: str
    source_paths: SourcePath

    # A path will be checked if it is covered by the allows list *and*
    # it is not not covered by the denylist.
    #
    # These paths can be regular files, not just directories.
    checked_directory_allowlist: Sequence[str] = dataclasses.field(default_factory=list)
    checked_directory_blocklist: Sequence[str] = dataclasses.field(default_factory=list)

    debug: bool = False
    enable_readonly_analysis: Optional[bool] = None
    enable_strict_override_check: Optional[bool] = None
    enable_strict_any_check: Optional[bool] = None
    enable_unawaited_awaitable_analysis: Optional[bool] = None
    excludes: Sequence[str] = dataclasses.field(default_factory=list)
    extensions: Sequence[str] = dataclasses.field(default_factory=list)
    include_suppressed_errors: Optional[bool] = None
    relative_local_root: Optional[str] = None
    memory_profiling_output: Optional[Path] = None
    number_of_workers: int = 1
    parallel: bool = True
    profiling_output: Optional[Path] = None
    python_version: configuration_module.PythonVersion = (
        configuration_module.PythonVersion(major=3)
    )
    system_platform: Optional[str] = None
    shared_memory: configuration_module.SharedMemory = (
        configuration_module.SharedMemory()
    )
    remote_logging: Optional[RemoteLogging] = None
    search_paths: Sequence[search_path.Element] = dataclasses.field(
        default_factory=list
    )

    def get_local_root(self) -> Optional[str]:
        if self.relative_local_root is None:
            return None
        return os.path.join(self.global_root, self.relative_local_root)

    def serialize(self) -> Dict[str, Any]:
        local_root = self.get_local_root()
        return {
            "source_paths": self.source_paths.serialize(),
            "search_paths": [
                element.command_line_argument() for element in self.search_paths
            ],
            "excludes": self.excludes,
            "checked_directory_allowlist": self.checked_directory_allowlist,
            "checked_directory_blocklist": self.checked_directory_blocklist,
            **(
                {"enable_readonly_analysis": self.enable_readonly_analysis}
                if self.enable_readonly_analysis is not None
                else {}
            ),
            **(
                {"enable_strict_override_check": self.enable_strict_override_check}
                if self.enable_strict_override_check is not None
                else {}
            ),
            **(
                {"enable_strict_any_check": self.enable_strict_any_check}
                if self.enable_strict_any_check is not None
                else {}
            ),
            **(
                {
                    "enable_unawaited_awaitable_analysis": (
                        self.enable_unawaited_awaitable_analysis
                    )
                }
                if self.enable_unawaited_awaitable_analysis is not None
                else {}
            ),
            "extensions": self.extensions,
            **(
                {"include_suppressed_errors": (self.include_suppressed_errors)}
                if self.include_suppressed_errors is not None
                else {}
            ),
            "log_path": self.log_path,
            "global_root": self.global_root,
            **({} if local_root is None else {"local_root": local_root}),
            "debug": self.debug,
            "python_version": {
                "major": self.python_version.major,
                "minor": self.python_version.minor,
                "micro": self.python_version.micro,
            },
            **(
                {}
                if self.system_platform is None
                else {"system_platform": self.system_platform}
            ),
            "shared_memory": self.shared_memory.to_json(),
            "parallel": self.parallel,
            "number_of_workers": self.number_of_workers,
            **(
                {}
                if self.remote_logging is None
                else {"remote_logging": self.remote_logging.serialize()}
            ),
            **(
                {}
                if self.profiling_output is None
                else {"profiling_output": str(self.profiling_output)}
            ),
            **(
                {}
                if self.memory_profiling_output is None
                else {"memory_profiling_output": str(self.memory_profiling_output)}
            ),
        }


def find_watchman_root(
    base: Path,
    stop_search_after: Optional[int] = None,
) -> Optional[Path]:
    return find_directories.find_parent_directory_containing_file(
        base, ".watchmanconfig", stop_search_after
    )


def find_buck_root(
    base: Path,
    stop_search_after: Optional[int] = None,
) -> Optional[Path]:
    return find_directories.find_parent_directory_containing_file(
        base, ".buckconfig", stop_search_after
    )


def find_buck2_root(
    base: Path,
    stop_search_after: Optional[int] = None,
) -> Optional[Path]:
    # Buck2 uses project root instead of cell root as its base directory.
    # This is essentially what `buck2 root --kind project` does.
    return find_directories.find_outermost_directory_containing_file(
        base, ".buckconfig", stop_search_after
    )


def _get_global_or_local_root(
    configuration: frontend_configuration.Base,
) -> Path:
    global_root = configuration.get_global_root()
    relative_local_root = configuration.get_relative_local_root()
    return (
        (global_root / relative_local_root)
        if relative_local_root is not None
        else global_root
    )


def get_source_path(
    configuration: frontend_configuration.Base,
    artifact_root_name: str,
    flavor: identifiers.PyreFlavor,
    kill_buck_after_build: bool,
    number_of_buck_threads: Optional[int],
    watchman_root: Optional[Path],
) -> SourcePath:
    source_directories = configuration.is_source_directories_defined()
    targets = configuration.get_buck_targets()
    buck_mode = configuration.get_buck_mode()

    if source_directories and targets is None:
        elements: Sequence[search_path.Element] = (
            configuration.get_existent_source_directories()
        )
        if len(elements) == 0:
            LOG.warning("Pyre did not find an existent source directory.")

        unwatched_dependency = configuration.get_existent_unwatched_dependency()
        if unwatched_dependency is not None and watchman_root is not None:
            return WithUnwatchedDependencySourcePath(
                change_indicator_root=watchman_root,
                unwatched_dependency=unwatched_dependency,
                elements=elements,
            )
        else:
            return SimpleSourcePath(elements)

    if (
        source_directories
        and targets is not None
        and flavor == identifiers.PyreFlavor.CLASSIC
    ):
        raise configuration_module.InvalidConfiguration(
            "`source_directories` and `targets` are mutually exclusive for typechecking"
        )

    if targets is not None:
        use_buck2 = configuration.uses_buck2()
        search_base = _get_global_or_local_root(configuration)
        source_root = (
            find_buck2_root(search_base) if use_buck2 else find_buck_root(search_base)
        )
        if source_root is None:
            raise configuration_module.InvalidConfiguration(
                "Cannot find a buck root for the specified targets. "
                + "Make sure the project is covered by a `.buckconfig` file."
            )
        source_directories = configuration.get_existent_source_directories()
        return BuckSourcePath(
            source_root=source_root,
            artifact_root=configuration.get_dot_pyre_directory() / artifact_root_name,
            checked_directory=search_base,
            targets=targets,
            targets_fallback_sources=(
                None if len(source_directories) == 0 else source_directories
            ),
            mode=buck_mode,
            isolation_prefix=configuration.get_buck_isolation_prefix(),
            bxl_builder=configuration.get_buck_bxl_builder(),
            use_buck2=use_buck2,
            kill_buck_after_build=kill_buck_after_build,
            number_of_threads=number_of_buck_threads,
        )

    raise configuration_module.InvalidConfiguration(
        "Cannot find any source files to analyze. "
        + "Either `source_directories` or `targets` must be specified."
    )


def get_source_path_for_server(
    configuration: frontend_configuration.Base,
    flavor: identifiers.PyreFlavor,
    kill_buck_after_build: bool,
    number_of_buck_threads: Optional[int],
    watchman_root: Optional[Path] = None,
) -> SourcePath:
    # We know that for each source root there could be at most one server alive.
    # Therefore artifact root name can be a fixed constant.
    artifact_root_name = f"{SERVER_ARTIFACT_ROOT_NAME}{flavor.path_suffix()}"
    relative_local_root = configuration.get_relative_local_root()
    if relative_local_root is not None:
        # Prevent artifact roots of different local projects from clashing with
        # each other.
        artifact_root_name = str(Path(artifact_root_name) / relative_local_root)
    return get_source_path(
        configuration,
        artifact_root_name,
        flavor,
        kill_buck_after_build,
        number_of_buck_threads,
        watchman_root,
    )


def get_source_path_for_check(
    configuration: frontend_configuration.Base,
    kill_buck_after_build: bool,
    number_of_buck_threads: Optional[int],
) -> SourcePath:
    # Artifact for one-off check command should not be a fixed constant, to prevent
    # concurrent check commands overwriting each other's artifacts. Here we use process
    # ID to isolate the artifact root of each individual check command.
    return get_source_path(
        configuration,
        str(os.getpid()),
        identifiers.PyreFlavor.CLASSIC,
        kill_buck_after_build,
        number_of_buck_threads,
        watchman_root=None,
    )


def get_checked_directory_allowlist(
    configuration: frontend_configuration.Base, source_path: SourcePath
) -> List[str]:
    source_path_allowlist = list(source_path.get_checked_directory_allowlist())
    explicit_allowlist = list(configuration.get_only_check_paths())
    # If allowlist paths were specifically provided, do not include inferred paths.
    return explicit_allowlist or source_path_allowlist


def get_profiling_log_path(log_directory: Path) -> Path:
    return log_directory / "profiling.log"


class SerializableArguments(Protocol):
    def serialize(self) -> Dict[str, Any]: ...


def _write_argument_file(
    output_file: IO[str], arguments: SerializableArguments
) -> None:
    LOG.info(f"Writing arguments into {output_file.name}...")
    serialized_arguments = arguments.serialize()
    LOG.debug(f"Arguments:\n{json.dumps(serialized_arguments, indent=2)}")
    output_file.write(json.dumps(serialized_arguments))
    output_file.flush()


@contextlib.contextmanager
def temporary_argument_file(arguments: SerializableArguments) -> Iterator[Path]:
    with tempfile.NamedTemporaryFile(
        mode="w", prefix="pyre_arguments_", suffix=".json"
    ) as argument_file:
        _write_argument_file(argument_file, arguments)
        yield Path(argument_file.name)


@dataclasses.dataclass
class LogFile:
    name: str
    file: IO[str]


@contextlib.contextmanager
def backend_log_file(prefix: str) -> Iterator[LogFile]:
    with tempfile.NamedTemporaryFile(
        mode="w", prefix=prefix, suffix=".log", delete=True
    ) as argument_file:
        yield LogFile(name=argument_file.name, file=argument_file.file)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This subpackage contains classes describing Pyre configurations, as
well as logic for finding and reading configuration files.
"""

from . import extension, search_path  # noqa: F401
from .configuration import (  # noqa: F401
    check_nested_local_configuration,
    Configuration,
    create_configuration,
    create_overridden_configuration,
    merge_partial_configurations,
    PartialConfiguration,
)
from .exceptions import InvalidConfiguration, InvalidPythonVersion  # noqa: F401
from .platform_aware import PlatformAware  # noqa: F401
from .python_version import PythonVersion  # noqa: F401
from .scheduler_policies import SchedulerPolicies  # noqa: F401
from .shared_memory import SharedMemory  # noqa: F401
from .unwatched import UnwatchedDependency, UnwatchedFiles  # noqa: F401



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module contains Python classes representing the information found
in a Pyre configuration file (.pyre_configuration).

The implementation is split into Base and OpenSource so that it is
possible to customize Pyre by implementing a new command-line tool
with additional configuration, using open-source Pyre as a library.
"""

from __future__ import annotations

import abc
import dataclasses
import json
import logging
import os
import shutil
import sys
from pathlib import Path
from typing import List, Optional

from . import configuration as configuration_module, find_directories, identifiers

LOG: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class SavedStateProject:
    name: str
    metadata: Optional[str] = None


class ServerStartCommand(abc.ABC):
    @abc.abstractmethod
    def get_pyre_binary_location(self) -> str: ...

    @abc.abstractmethod
    def get_start_command(
        self, argument_file_path: Path, flavor: identifiers.PyreFlavor
    ) -> List[str]: ...


class DefaultServerStartCommand(ServerStartCommand):
    def __init__(self, binary_location: str) -> None:
        self._binary_location = binary_location

    def __eq__(self, other: DefaultServerStartCommand) -> bool:
        return self._binary_location == other._binary_location

    def get_pyre_binary_location(self) -> str:
        return self._binary_location

    def get_start_command(
        self, argument_file_path: Path, flavor: identifiers.PyreFlavor
    ) -> List[str]:
        return [self._binary_location, "server", str(argument_file_path)]


# TODO(T120824066): Break this class down into smaller pieces. Ideally, one
# class per command.
class Base(abc.ABC):
    @abc.abstractmethod
    def get_dot_pyre_directory(self) -> Path:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_server_start_command(
        self, download_if_needed: bool = False
    ) -> Optional[ServerStartCommand]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_typeshed_location(self, download_if_needed: bool = False) -> Optional[Path]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_binary_version(self) -> Optional[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_content_for_display(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_global_root(self) -> Path:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_relative_local_root(self) -> Optional[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_excludes(self) -> List[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def is_strict(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_remote_logger(self) -> Optional[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_number_of_workers(self) -> int:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_python_version(self) -> configuration_module.PythonVersion:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_system_platform(self) -> Optional[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_shared_memory(self) -> configuration_module.SharedMemory:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_valid_extension_suffixes(self) -> List[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_ignore_all_errors(self) -> List[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_only_check_paths(self) -> List[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_existent_user_specified_search_paths(
        self,
    ) -> List[configuration_module.search_path.Element]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_existent_source_directories(
        self,
    ) -> List[configuration_module.search_path.Element]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_existent_unwatched_dependency(
        self,
    ) -> Optional[configuration_module.unwatched.UnwatchedDependency]:
        raise NotImplementedError()

    @abc.abstractmethod
    def is_source_directories_defined(
        self,
    ) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_buck_targets(
        self,
    ) -> Optional[List[str]]:
        raise NotImplementedError()

    @abc.abstractmethod
    def uses_buck2(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_buck_mode(self) -> Optional[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_buck_isolation_prefix(self) -> Optional[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_buck_bxl_builder(self) -> Optional[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_other_critical_files(self) -> List[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_taint_models_path(self) -> List[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_project_identifier(self) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_enable_readonly_analysis(self) -> Optional[bool]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_enable_strict_override_check(self) -> Optional[bool]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_enable_strict_any_check(self) -> Optional[bool]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_enable_unawaited_awaitable_analysis(self) -> Optional[bool]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_saved_state_project(self) -> Optional[SavedStateProject]:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_include_suppressed_errors(self) -> Optional[bool]:
        raise NotImplementedError()

    def get_local_root(self) -> Optional[Path]:
        relative_local_root = self.get_relative_local_root()
        if relative_local_root is None:
            return None
        return self.get_global_root() / relative_local_root

    def get_log_directory(self) -> Path:
        dot_pyre_directory = self.get_dot_pyre_directory()
        relative_local_root = self.get_relative_local_root()
        return (
            dot_pyre_directory
            if relative_local_root is None
            else dot_pyre_directory / relative_local_root
        )

    def get_existent_typeshed_search_paths(
        self,
    ) -> List[configuration_module.search_path.Element]:
        typeshed_root = self.get_typeshed_location(download_if_needed=True)
        if typeshed_root is None:
            return []
        return [
            configuration_module.search_path.SimpleElement(str(element))
            for element in find_directories.find_typeshed_search_paths(typeshed_root)
        ]

    def get_existent_search_paths(
        self,
    ) -> List[configuration_module.search_path.Element]:
        return [
            *self.get_existent_user_specified_search_paths(),
            *self.get_existent_typeshed_search_paths(),
        ]


class OpenSource(Base):
    def __init__(
        self,
        configuration: configuration_module.Configuration,
    ) -> None:
        self.configuration = configuration

    def get_dot_pyre_directory(self) -> Path:
        return (
            self.configuration.dot_pyre_directory
            or self.get_global_root() / find_directories.LOG_DIRECTORY
        )

    def _get_binary_location(self, download_if_needed: bool = False) -> Optional[Path]:
        binary = self.configuration.binary
        if binary is not None:
            return Path(binary)

        LOG.info(
            f"No binary specified, looking for `{find_directories.BINARY_NAME}` in PATH"
        )
        binary_candidate = shutil.which(find_directories.BINARY_NAME)
        if binary_candidate is None:
            binary_candidate_name = os.path.join(
                os.path.dirname(sys.argv[0]), find_directories.BINARY_NAME
            )
            binary_candidate = shutil.which(binary_candidate_name)

        # Auto-download is not supported in OSS
        return Path(binary_candidate) if binary_candidate is not None else None

    def get_server_start_command(
        self, download_if_needed: bool = False
    ) -> Optional[ServerStartCommand]:
        binary_location = self._get_binary_location(download_if_needed)
        if binary_location is not None:
            return DefaultServerStartCommand(str(binary_location))
        return None

    def get_typeshed_location(self, download_if_needed: bool = False) -> Optional[Path]:
        typeshed = self.configuration.typeshed
        if typeshed is not None:
            return Path(typeshed)

        LOG.info("No typeshed specified, looking for it...")
        auto_determined_typeshed = find_directories.find_typeshed()
        if auto_determined_typeshed is None:
            # Auto-download is not supported in OSS
            LOG.warning(
                "Could not find a suitable typeshed. Types for Python builtins "
                "and standard libraries may be missing!"
            )
            return None
        else:
            LOG.info(f"Found: `{auto_determined_typeshed}`")
            return auto_determined_typeshed

    def get_binary_version(self) -> Optional[str]:
        return None

    def get_content_for_display(self) -> str:
        return json.dumps(self.configuration.to_json(), indent=2)

    def get_global_root(self) -> Path:
        return self.configuration.global_root

    def get_relative_local_root(self) -> Optional[str]:
        return self.configuration.relative_local_root

    def get_excludes(self) -> List[str]:
        return list(self.configuration.excludes)

    def is_strict(self) -> bool:
        return self.configuration.strict

    def get_remote_logger(self) -> Optional[str]:
        return self.configuration.logger

    def get_number_of_workers(self) -> int:
        return self.configuration.get_number_of_workers()

    def get_python_version(self) -> configuration_module.PythonVersion:
        return self.configuration.get_python_version()

    def get_system_platform(self) -> Optional[str]:
        return self.configuration.get_system_platform()

    def get_shared_memory(self) -> configuration_module.SharedMemory:
        return self.configuration.shared_memory

    def get_valid_extension_suffixes(self) -> List[str]:
        return self.configuration.get_valid_extension_suffixes()

    def get_ignore_all_errors(self) -> List[str]:
        return list(self.configuration.ignore_all_errors)

    def get_only_check_paths(self) -> List[str]:
        return list(self.configuration.only_check_paths)

    def get_existent_user_specified_search_paths(
        self,
    ) -> List[configuration_module.search_path.Element]:
        return self.configuration.expand_and_get_existent_search_paths()

    def get_existent_source_directories(
        self,
    ) -> List[configuration_module.search_path.Element]:
        return self.configuration.expand_and_get_existent_source_directories()

    def get_existent_unwatched_dependency(
        self,
    ) -> Optional[configuration_module.unwatched.UnwatchedDependency]:
        return self.configuration.get_existent_unwatched_dependency()

    def is_source_directories_defined(
        self,
    ) -> bool:
        return self.configuration.source_directories is not None

    def get_buck_targets(
        self,
    ) -> Optional[List[str]]:
        targets = self.configuration.targets
        return list(targets) if targets is not None else None

    def uses_buck2(self) -> bool:
        return self.configuration.use_buck2

    def get_buck_mode(self) -> Optional[str]:
        mode = self.configuration.buck_mode
        return mode.get() if mode is not None else None

    def get_buck_isolation_prefix(self) -> Optional[str]:
        return self.configuration.isolation_prefix

    def get_buck_bxl_builder(self) -> Optional[str]:
        return self.configuration.bxl_builder

    def get_other_critical_files(self) -> List[str]:
        return list(self.configuration.other_critical_files)

    def get_taint_models_path(self) -> List[str]:
        return list(self.configuration.taint_models_path)

    def get_enable_readonly_analysis(self) -> Optional[bool]:
        return self.configuration.enable_readonly_analysis

    def get_enable_strict_override_check(self) -> Optional[bool]:
        return self.configuration.enable_strict_override_check

    def get_enable_strict_any_check(self) -> Optional[bool]:
        return self.configuration.enable_strict_any_check

    def get_enable_unawaited_awaitable_analysis(self) -> Optional[bool]:
        return self.configuration.enable_unawaited_awaitable_analysis

    def get_project_identifier(self) -> str:
        return self.configuration.project_identifier

    def get_saved_state_project(self) -> Optional[SavedStateProject]:
        return None

    def get_include_suppressed_errors(self) -> Optional[bool]:
        return self.configuration.include_suppressed_errors



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module contains filesystem-related utility code including:
- path expansion tools
"""

from pathlib import Path


def expand_relative_path(root: str, path: str) -> str:
    expanded_path = Path(path).expanduser()
    if expanded_path.is_absolute():
        return str(expanded_path)
    else:
        return str(Path(root) / expanded_path)


def expand_global_root(path: str, global_root: str) -> str:
    if path.startswith("//"):
        return expand_relative_path(global_root, path[2:])
    return path



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module specifies how to configure the Python version in Pyre.
"""

import dataclasses

from . import exceptions


@dataclasses.dataclass(frozen=True)
class PythonVersion:
    major: int
    minor: int = 0
    micro: int = 0

    @staticmethod
    def from_string(input: str) -> "PythonVersion":
        try:
            splits = input.split(".")
            if len(splits) == 1:
                return PythonVersion(major=int(splits[0]))
            elif len(splits) == 2:
                return PythonVersion(major=int(splits[0]), minor=int(splits[1]))
            elif len(splits) == 3:
                return PythonVersion(
                    major=int(splits[0]), minor=int(splits[1]), micro=int(splits[2])
                )
            raise exceptions.InvalidPythonVersion(
                "Version string is expected to have the form of 'X.Y.Z' but got "
                + f"'{input}'"
            )
        except ValueError as error:
            raise exceptions.InvalidPythonVersion(input) from error

    def to_string(self) -> str:
        return f"{self.major}.{self.minor}.{self.micro}"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
Shared logic for client-side handling of subscriptions to the
Pyre daemon. Context on this:
- A Pyre daemon opens a socket accepting client connections.
- The socket accepts one-off requests using a request/response model
- But the socket also allows long-running connections where we send a
  subscription request, and the connection stays open. The server will
  then push notifications (type errors and/or state change alerts) to
  the client.
"""

import dataclasses
import json
from typing import Optional, Union

from . import incremental

TypeErrors = incremental.TypeErrors


def _parse_type_error_subscription(response: object) -> TypeErrors:
    return incremental.parse_type_error_response_json(["TypeErrors", response])


@dataclasses.dataclass(frozen=True)
class StatusUpdate:
    kind: str
    message: Optional[str] = None


def _parse_status_update_subscription(response: object) -> StatusUpdate:
    if not isinstance(response, list) or len(response) == 0:
        raise incremental.InvalidServerResponse(
            f"Status update subscription must be a nonempty list. Got {response}"
        )
    kind = response[0]
    if not isinstance(kind, str):
        raise incremental.InvalidServerResponse(
            f"Response kind of a status update must be a string. Got {response}"
        )
    message = None
    if len(response) > 1 and isinstance(response[1], dict) and "message" in response[1]:
        message = response[1]["message"]
    return StatusUpdate(kind=kind, message=message)


@dataclasses.dataclass(frozen=True)
class Error:
    message: str


def _parse_error_subscription(response: object) -> Error:
    if not isinstance(response, str):
        raise incremental.InvalidServerResponse(
            f"Response kind of an error must be a string. Got {response}"
        )
    return Error(message=response)


class IncrementalTelemetry:
    pass


Body = Union[TypeErrors, StatusUpdate, Error, IncrementalTelemetry]


@dataclasses.dataclass(frozen=True)
class Response:
    body: Body

    @staticmethod
    def parse(response: str) -> "Response":
        try:
            response_json = json.loads(response)
            # The response JSON is expected to have the following forms:
            # `{"name": "foo", "body": ["TypeErrors", [error_json, ...]]}`
            # `{"name": "foo", "body": ["StatusUpdate", ["message_kind", ...]]}`
            # `{"name": "foo", "body": ["Error", "error message"]}`
            # The legacy subscription protocol includes a subscription name - we have
            # since moved away from requiring a name associated with a subscription, as
            # there aren't practical ways to set up multiple subscriptions on the same connection,
            # and a connection serves as a unique identifier for a subscription.

            # For backwards compatibility, we allow the "name" field to be in the response, but do not
            # parse it on the client.
            if isinstance(response_json, dict):
                body = response_json.get("body", None)
                if body is not None and isinstance(body, list) and len(body) > 1:
                    tag = body[0]
                    if tag == "TypeErrors":
                        return Response(body=_parse_type_error_subscription(body[1]))
                    elif tag == "StatusUpdate":
                        return Response(body=_parse_status_update_subscription(body[1]))
                    elif tag == "Error":
                        return Response(body=_parse_error_subscription(body[1]))
                    elif tag == "IncrementalTelemetry":
                        return Response(body=IncrementalTelemetry())

            raise incremental.InvalidServerResponse(
                f"Unexpected JSON subscription from server: {response_json}"
            )
        except json.JSONDecodeError as decode_error:
            message = f"Cannot parse subscription as JSON: {decode_error}"
            raise incremental.InvalidServerResponse(message) from decode_error



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module is responsible for handling requests from the VScode language server and generating an appropriate response.

The response typically will be generated through the Pyre daemon, and the name PyreLanguageServer was chosen for this module
because it illustrates that this is the intermediary between the Language server and the Pyre daemon.
"""

from __future__ import annotations

import abc
import asyncio
import dataclasses
import logging
from pathlib import Path
from typing import (
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
)

from .. import (
    background_tasks,
    error,
    json_rpc,
    log,
    timer,
)
from ..language_server import connections, daemon_connection, features, protocol as lsp
from . import (
    commands,
    daemon_querier,
    server_state as state,
    type_error_handler,
)
from .daemon_querier import DaemonQueryFailure
from .server_state import OpenedDocumentState

LOG: logging.Logger = logging.getLogger(__name__)
CONSECUTIVE_START_ATTEMPT_THRESHOLD: int = 5


async def read_lsp_request(
    input_channel: connections.AsyncTextReader,
    output_channel: connections.AsyncTextWriter,
) -> json_rpc.Request:
    while True:
        try:
            message = await lsp.read_json_rpc(input_channel)
            return message
        except json_rpc.JSONRPCException as json_rpc_error:
            LOG.error(f"Exception occurred while reading JSON RPC: {json_rpc_error}")
            await lsp.write_json_rpc_ignore_connection_error(
                output_channel,
                json_rpc.ErrorResponse(
                    id=None,
                    code=json_rpc_error.error_code(),
                    message=str(json_rpc_error),
                ),
            )


async def _wait_for_exit(
    input_channel: connections.AsyncTextReader,
    output_channel: connections.AsyncTextWriter,
) -> None:
    """
    Wait for an LSP "exit" request from the `input_channel`. This is mostly useful
    when the LSP server has received a "shutdown" request, in which case the LSP
    specification dictates that only "exit" can be sent from the client side.

    If a non-exit LSP request is received, drop it and keep waiting on another
    "exit" request.
    """
    while True:
        request = await read_lsp_request(input_channel, output_channel)
        if request.method != "exit":
            LOG.debug(f"Non-exit request received after shutdown: {request}")
            continue
        # Got an exit request. Stop the wait.
        return


def daemon_failure_string(operation: str, type_string: str, error_message: str) -> str:
    return f"For {operation} request, encountered failure response of type: {type_string}, error_message: {error_message}"


class PyreLanguageServerApi(abc.ABC):
    @abc.abstractmethod
    async def write_telemetry(
        self,
        parameters: Dict[str, object],
        activity_key: Optional[Dict[str, object]],
    ) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_language_server_features(self) -> features.LanguageServerFeatures:
        raise NotImplementedError()

    @abc.abstractmethod
    async def update_overlay_if_needed(self, document_path: Path) -> float:
        raise NotImplementedError()

    @abc.abstractmethod
    async def process_open_request(
        self,
        parameters: lsp.DidOpenTextDocumentParameters,
        activity_key: Optional[Dict[str, object]] = None,
    ) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def process_close_request(
        self, parameters: lsp.DidCloseTextDocumentParameters
    ) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def process_did_change_request(
        self,
        parameters: lsp.DidChangeTextDocumentParameters,
        activity_key: Optional[Dict[str, object]] = None,
    ) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def process_did_save_request(
        self,
        parameters: lsp.DidSaveTextDocumentParameters,
        activity_key: Optional[Dict[str, object]] = None,
    ) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def process_type_coverage_request(
        self,
        parameters: lsp.TypeCoverageParameters,
        request_id: Union[int, str, None],
        activity_key: Optional[Dict[str, object]] = None,
    ) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def process_shutdown_request(self, request_id: Union[int, str, None]) -> None:
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class PyreLanguageServer(PyreLanguageServerApi):
    # Channel to send responses to the editor
    output_channel: connections.AsyncTextWriter

    # NOTE: The fields inside `server_state` are mutable and can be changed by the background
    # task.
    server_state: state.ServerState

    querier: daemon_querier.AbstractDaemonQuerier
    client_type_error_handler: type_error_handler.ClientTypeErrorHandler

    async def write_telemetry(
        self,
        parameters: Dict[str, object],
        activity_key: Optional[Dict[str, object]],
    ) -> None:
        should_write_telemetry = self.server_state.server_options.language_server_features.telemetry.is_enabled()
        if should_write_telemetry:
            parameters = dict(parameters)
            parameters["project_identifier"] = (
                self.server_state.server_options.project_identifier
            )
            await lsp.write_json_rpc_ignore_connection_error(
                self.output_channel,
                json_rpc.Request(
                    activity_key=activity_key,
                    method="telemetry/event",
                    parameters=json_rpc.ByNameParameters(parameters),
                ),
            )

    def get_language_server_features(self) -> features.LanguageServerFeatures:
        return self.server_state.server_options.language_server_features

    async def update_overlay_if_needed(self, document_path: Path) -> float:
        """
        Send an overlay update to the daemon if three conditions are met:
        - unsaved changes support is enabled
        - a document is listed in `server_state.opened_documents`
        - the OpenedDocumentState says the overlay overlay may be stale

        Returns the time taken to run the update.
        """
        update_timer = timer.Timer()
        if (
            self.get_language_server_features().unsaved_changes.is_enabled()
            and document_path in self.server_state.opened_documents
        ):
            opened_document_state = self.server_state.opened_documents[document_path]
            code_changes = opened_document_state.code
            current_is_dirty_state = opened_document_state.is_dirty
            if not opened_document_state.pyre_code_updated:
                result = await self.querier.update_overlay(
                    path=document_path, code=code_changes
                )
                if isinstance(result, daemon_connection.DaemonConnectionFailure):
                    LOG.info(
                        daemon_failure_string(
                            "didChange", str(type(result)), result.error_message
                        )
                    )
                    LOG.info(result.error_message)
                else:
                    self.server_state.opened_documents[document_path] = (
                        OpenedDocumentState(
                            code=code_changes,
                            is_dirty=current_is_dirty_state,
                            pyre_code_updated=True,
                        )
                    )
        return update_timer.stop_in_millisecond()

    async def process_open_request(
        self,
        parameters: lsp.DidOpenTextDocumentParameters,
        activity_key: Optional[Dict[str, object]] = None,
    ) -> None:
        document_path = parameters.text_document.document_uri().to_file_path()
        if document_path is None:
            raise json_rpc.InvalidRequestError(
                f"Document URI is not a file: {parameters.text_document.uri}"
            )
        document_path = document_path.resolve()
        self.server_state.opened_documents[document_path] = OpenedDocumentState(
            code=parameters.text_document.text,
            is_dirty=False,
            pyre_code_updated=True,
        )
        LOG.info(f"File opened: {document_path}")
        await self.querier.handle_file_opened(
            document_path, parameters.text_document.text
        )

    async def process_close_request(
        self, parameters: lsp.DidCloseTextDocumentParameters
    ) -> None:
        document_path = parameters.text_document.document_uri().to_file_path()
        if document_path is None:
            raise json_rpc.InvalidRequestError(
                f"Document URI is not a file: {parameters.text_document.uri}"
            )
        document_path = document_path.resolve()
        try:
            del self.server_state.opened_documents[document_path]
            LOG.info(f"File closed: {document_path}")
            await self.querier.handle_file_closed(document_path)
        except KeyError:
            LOG.warning(f"Trying to close an un-opened file: {document_path}")

    async def _publish_type_errors_for_files(
        self,
        type_errors: Mapping[Path, Collection[error.Error]],
        type_checked_files: Set[Path],
        set_unused_as_warning: bool = False,
    ) -> None:
        for file in type_checked_files:
            document_errors = type_errors.get(file, set())
            await self.client_type_error_handler.show_overlay_type_errors(
                path=file,
                type_errors=list(document_errors),
                set_unused_as_warning=set_unused_as_warning,
            )

    async def _query_pyre_daemon_type_errors(
        self, document_path: Path, type_checkable_files: Set[Path]
    ) -> Tuple[Dict[Path, List[error.Error]], Optional[str]]:
        if len(type_checkable_files) == 0:
            LOG.debug("No daemon type checkable files found")
            return {}, None
        await self.update_overlay_if_needed(document_path)
        result = await self.querier.get_type_errors(
            type_checkable_files,
        )
        type_errors: Dict[Path, List[error.Error]] = {}
        if isinstance(result, DaemonQueryFailure):
            return type_errors, result.error_message
        else:
            type_errors = result
            await self._publish_type_errors_for_files(type_errors, type_checkable_files)
            return type_errors, None

    async def handle_overlay_type_errors(
        self,
        document_path: Path,
        new_file_loaded: bool,
        activity_key: Optional[Dict[str, object]] = None,
    ) -> None:
        client_register_event = self.server_state.client_register_event
        if client_register_event is not None and not client_register_event.is_set():
            return
        daemon_status_before = self.server_state.status_tracker.get_status()
        type_errors_timer = timer.Timer()
        open_documents = set(self.server_state.opened_documents.keys())
        type_errors, error_message = await self._query_pyre_daemon_type_errors(
            document_path, open_documents
        )

        json_type_errors = {
            str(path): [type_error.to_json() for type_error in errors]
            for path, errors in type_errors.items()
        }
        await self.write_telemetry(
            {
                "type": "LSP",
                "operation": "typeErrors",
                "filePath": str(document_path),
                "server_state_open_documents_count": len(
                    self.server_state.opened_documents
                ),
                "duration_ms": type_errors_timer.stop_in_millisecond(),
                "error_message": error_message,
                "type_errors": json_type_errors,
                **daemon_status_before.as_telemetry_dict(),
            },
            activity_key,
        )

    async def process_did_change_request(
        self,
        parameters: lsp.DidChangeTextDocumentParameters,
        activity_key: Optional[Dict[str, object]] = None,
    ) -> None:
        document_path = parameters.text_document.document_uri().to_file_path()
        if document_path is None:
            raise json_rpc.InvalidRequestError(
                f"Document URI is not a file: {parameters.text_document.uri}"
            )
        document_path = document_path.resolve()
        if document_path not in self.server_state.opened_documents:
            return

        daemon_status_before = self.server_state.status_tracker.get_status()
        did_change_timer = timer.Timer()

        process_unsaved_changes = self.server_state.server_options.language_server_features.unsaved_changes.is_enabled()
        error_message = None
        code_changes = str(
            "".join(
                [content_change.text for content_change in parameters.content_changes]
            )
        )
        self.server_state.opened_documents[document_path] = OpenedDocumentState(
            code=code_changes,
            is_dirty=True,
            pyre_code_updated=False,
        )
        await self.write_telemetry(
            {
                "type": "LSP",
                "operation": "didChange",
                "filePath": str(document_path),
                "server_state_open_documents_count": len(
                    self.server_state.opened_documents
                ),
                "duration_ms": did_change_timer.stop_in_millisecond(),
                "error_message": error_message,
                "overlays_enabled": process_unsaved_changes,
                **daemon_status_before.as_telemetry_dict(),
            },
            activity_key,
        )
        if (
            process_unsaved_changes
            and self.get_language_server_features().type_errors.is_enabled()
        ):
            await self.handle_overlay_type_errors(
                document_path=document_path,
                new_file_loaded=False,
                activity_key=activity_key,
            )

    async def process_did_save_request(
        self,
        parameters: lsp.DidSaveTextDocumentParameters,
        activity_key: Optional[Dict[str, object]] = None,
    ) -> None:
        document_path = parameters.text_document.document_uri().to_file_path()
        if document_path is None:
            raise json_rpc.InvalidRequestError(
                f"Document URI is not a file: {parameters.text_document.uri}"
            )
        document_path = document_path.resolve()

        if document_path not in self.server_state.opened_documents:
            return

        daemon_status_before = self.server_state.status_tracker.get_status()

        code_changes = self.server_state.opened_documents[document_path].code

        self.server_state.opened_documents[document_path] = OpenedDocumentState(
            code=code_changes,
            is_dirty=False,
            # False here because even though a didSave event means the base environment
            # will be up-to-date (after an incremental push), it is not necessarily
            # the case that the overlay environment is up to date.
            pyre_code_updated=False,
        )
        await self.write_telemetry(
            {
                "type": "LSP",
                "operation": "didSave",
                "filePath": str(document_path),
                "server_state_open_documents_count": len(
                    self.server_state.opened_documents
                ),
                # We don't do any blocking work on didSave, but analytics are easier if
                # we avoid needlessly introducing NULL values.
                "duration_ms": 0,
                **daemon_status_before.as_telemetry_dict(),
            },
            activity_key,
        )

    async def process_type_coverage_request(
        self,
        parameters: lsp.TypeCoverageParameters,
        request_id: Union[int, str, None],
        activity_key: Optional[Dict[str, object]] = None,
    ) -> None:
        document_path = parameters.text_document.document_uri().to_file_path()
        if document_path is None:
            raise json_rpc.InvalidRequestError(
                f"Document URI is not a file: {parameters.text_document.uri}"
            )
        document_path = document_path.resolve()

        daemon_status_before = self.server_state.status_tracker.get_status()
        type_coverage_timer = timer.Timer()

        response = await self.querier.get_type_coverage(path=document_path)
        if response is not None:
            await lsp.write_json_rpc(
                self.output_channel,
                json_rpc.SuccessResponse(
                    id=request_id,
                    activity_key=activity_key,
                    result=response.to_dict(),
                ),
            )
        await self.write_telemetry(
            {
                "type": "LSP",
                "operation": "typeCoverage",
                "filePath": str(document_path),
                "duration_ms": type_coverage_timer.stop_in_millisecond(),
                "coverage_type": self.get_language_server_features().type_coverage.value,
                **daemon_status_before.as_telemetry_dict(),
            },
            activity_key,
        )

    async def process_shutdown_request(self, request_id: Union[int, str, None]) -> None:
        await lsp.write_json_rpc_ignore_connection_error(
            self.output_channel,
            json_rpc.SuccessResponse(id=request_id, activity_key=None, result=None),
        )


class PyreLanguageServerDispatcher:
    """
    The dispatcher provides the top-level, "foreground" logic for a Pyre
    language server. Its only job is to read requests from standard input,
    parse them, and dispatch to the appropriate lower-level logic.

    There are two compontents to which we might dispatch:
    - We'll dispatch to the PyreLanguageServer for all request handling,
      which includes querying the daemon, sending responses to the client,
      and reporting telemetry.
    - We also may check that the background task used to start/restart the
      daemon and get type error notifications over subscriptions is alive.
      The daemon can go down, for example if a critical file change occurs,
      so it is important for us to periodically check whether it is up.
    """

    # I/O channels. Output channel is used *exclusively* to report parse errors.
    input_channel: connections.AsyncTextReader
    output_channel: connections.AsyncTextWriter

    # State: used *exclusively* to track restart failures.
    server_state: state.ServerState

    daemon_manager: background_tasks.TaskManager
    api: PyreLanguageServerApi

    # A set of outstanding (not "done") asyncio tasks (like requests being processed). This is necessary to retain strong references to those tasks
    # to avoid them being collected mid-execution by gc. See https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
    outstanding_tasks: Set[asyncio.Task[None]]

    def __init__(
        self,
        input_channel: connections.AsyncTextReader,
        output_channel: connections.AsyncTextWriter,
        server_state: state.ServerState,
        daemon_manager: background_tasks.TaskManager,
        api: PyreLanguageServerApi,
    ) -> None:
        self.input_channel = input_channel
        self.output_channel = output_channel
        self.server_state = server_state
        self.daemon_manager = daemon_manager
        self.api = api
        self.outstanding_tasks = set()

    async def wait_for_exit(self) -> commands.ExitCode:
        await _wait_for_exit(self.input_channel, self.output_channel)
        return commands.ExitCode.SUCCESS

    async def _restart_if_needed(
        self, error_source: Optional[Exception] = None
    ) -> None:
        if (
            self.server_state.consecutive_start_failure
            >= CONSECUTIVE_START_ATTEMPT_THRESHOLD
        ):
            LOG.info(
                "Not restarting Pyre since failed consecutive start attempt limit"
                " has been reached."
            )
            return

        if isinstance(
            error_source,
            (
                connections.ConnectionFailure,
                asyncio.IncompleteReadError,
                ConnectionError,
            ),
        ):  # do we think the daemon is probably down at this point?
            # Terminate any existing daemon processes before starting a new one
            LOG.info("Forcing pyre daemon restart...")
            await self.daemon_manager.ensure_task_stop()  # make sure it's down

        # restart if needed
        if not self.daemon_manager.is_task_running():
            # Just check to ensure that the daemon is running and restart if needed
            await self.daemon_manager.ensure_task_running()

    async def dispatch_nonblocking_request(self, request: json_rpc.Request) -> None:
        if request.method == "exit" or request.method == "shutdown":
            raise Exception("Exit and shutdown requests should be blocking")

        dispatch_request_timer = timer.Timer()
        await self._restart_if_needed()
        dispatch_request_duration = (  # noqa: F841
            dispatch_request_timer.stop_in_millisecond()
        )
        if self.server_state.client_register_event is not None:
            await self.server_state.client_register_event.wait()
        elif request.method == "textDocument/didOpen":
            await self.api.process_open_request(
                lsp.DidOpenTextDocumentParameters.from_json_rpc_parameters(
                    request.extract_parameters()
                ),
                request.activity_key,
            )
            await self._restart_if_needed()
        elif request.method == "textDocument/didChange":
            await self.api.process_did_change_request(
                lsp.DidChangeTextDocumentParameters.from_json_rpc_parameters(
                    request.extract_parameters()
                )
            )
            await self._restart_if_needed()
        elif request.method == "textDocument/didClose":
            await self.api.process_close_request(
                lsp.DidCloseTextDocumentParameters.from_json_rpc_parameters(
                    request.extract_parameters()
                )
            )
        elif request.method == "textDocument/didSave":
            await self.api.process_did_save_request(
                lsp.DidSaveTextDocumentParameters.from_json_rpc_parameters(
                    request.extract_parameters()
                ),
                request.activity_key,
            )
            await self._restart_if_needed()
        elif request.method == "textDocument/typeCoverage":
            await self.api.process_type_coverage_request(
                lsp.TypeCoverageParameters.from_json_rpc_parameters(
                    request.extract_parameters()
                ),
                request.id,
                request.activity_key,
            )
        elif request.id is not None:
            raise lsp.RequestCancelledError(
                f"{request.method} Request not supported yet"
            )

    async def dispatch_request(
        self, request: json_rpc.Request
    ) -> Optional[commands.ExitCode]:
        """
        The top-level request dispatcher has two parts:
        - Forward the request to the appropriate handler method
        - For some types of requests, check that the background task is running; this
          is how we ensure the daemon connection is live (the background task will
          crash if the daemon goes down and closes the socket).

        """
        if request.method == "exit":
            LOG.info(
                "Received exit request without a shutdown request, exiting as FAILURE."
            )
            return commands.ExitCode.LANGUAGE_SERVER_EXIT
        elif request.method == "shutdown":
            await self.api.process_shutdown_request(request.id)
            return await self.wait_for_exit()
        else:
            request_task = asyncio.create_task(
                self.dispatch_nonblocking_request(request)
            )
            self.outstanding_tasks.add(request_task)
            request_task.add_done_callback(self.outstanding_tasks.discard)

    async def serve_requests(self) -> int:
        while True:
            request = await read_lsp_request(self.input_channel, self.output_channel)
            LOG.debug(f"Received LSP request: {log.truncate(str(request), 400)}")

            try:
                return_code = await self.dispatch_request(request)
                if return_code is not None:
                    return return_code
            except json_rpc.JSONRPCException as json_rpc_error:
                LOG.error(
                    f"Exception occurred while processing request: {json_rpc_error}"
                )
                await lsp.write_json_rpc_ignore_connection_error(
                    self.output_channel,
                    json_rpc.ErrorResponse(
                        id=request.id,
                        activity_key=request.activity_key,
                        code=json_rpc_error.error_code(),
                        message=str(json_rpc_error),
                    ),
                )

    async def run(self) -> int:
        """
        Launch the background tasks that deal with starting and subscribing
        to a pyre server and managing a queue of requests, then run the
        language server itself.
        """
        try:
            await self.daemon_manager.ensure_task_running()
            return await self.serve_requests()
        except lsp.ReadChannelClosedError:
            # This error can happen when the connection gets closed unilaterally
            # from the language client, which causes issue when we try to access the
            # input channel. This usually signals that the language client has exited,
            # which implies that the language server should do that as well.
            LOG.info("Connection closed by LSP client.")
            return commands.ExitCode.SUCCESS
        finally:
            await self.daemon_manager.ensure_task_stop()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides the logic for the `pyre servers` command, which lists
running servers (i.e. daemons - background backend processes available for
incremental checks and queries, including from language servers).
"""

import dataclasses
import json
import logging
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Sequence, Union

import tabulate
from typing_extensions import TypedDict

from .. import command_arguments, daemon_socket, identifiers, log
from ..language_server import connections
from . import commands, stop

LOG: logging.Logger = logging.getLogger(__name__)


class _ServerConfigurationJSONSchema(TypedDict):
    global_root: str


class _ServerInfoJSONSchema(TypedDict):
    pid: int
    version: str
    configuration: _ServerConfigurationJSONSchema


class InvalidServerResponse(Exception):
    pass


@dataclasses.dataclass(frozen=True)
class RunningServerStatus:
    pid: int
    version: str
    global_root: str
    flavor: str
    relative_local_root: Optional[str] = None

    @staticmethod
    def from_json(
        input_json: Dict[str, object], flavor: identifiers.PyreFlavor
    ) -> "RunningServerStatus":
        pid = input_json.get("pid", None)
        if not isinstance(pid, int):
            raise InvalidServerResponse(f"Expect `pid` to be an int but got {pid}")
        version = input_json.get("version", None)
        if not isinstance(version, str):
            raise InvalidServerResponse(
                f"Expect `version` to be a string but got {version}"
            )
        global_root = input_json.get("global_root", None)
        if not isinstance(global_root, str):
            raise InvalidServerResponse(
                f"Expect `global_root` to be a string but got {global_root}"
            )
        relative_local_root = input_json.get("relative_local_root", None)
        if relative_local_root is not None and not isinstance(relative_local_root, str):
            raise InvalidServerResponse(
                "Expected `relative_local_root` to be a string but got "
                f"{relative_local_root}"
            )
        return RunningServerStatus(
            pid=pid,
            version=version,
            global_root=global_root,
            relative_local_root=relative_local_root,
            flavor=flavor.value,
        )

    @staticmethod
    def from_server_response(
        response: str, flavor: identifiers.PyreFlavor
    ) -> "RunningServerStatus":
        try:
            response_json = json.loads(response)
            if (
                not isinstance(response_json, list)
                or len(response_json) < 2
                or response_json[0] != "Info"
                or not isinstance(response_json[1], dict)
            ):
                message = f"Unexpected JSON response: {response_json}"
                raise InvalidServerResponse(message)
            return RunningServerStatus.from_json(response_json[1], flavor)
        except json.JSONDecodeError as error:
            message = f"Cannot parse response as JSON: {error}"
            raise InvalidServerResponse(message) from error

    def to_json(self) -> Dict[str, object]:
        return {
            "status": "running",
            "pid": self.pid,
            "version": self.version,
            "global_root": self.global_root,
            "relative_local_root": self.relative_local_root,
            "flavor": self.flavor,
        }


@dataclasses.dataclass(frozen=True)
class DefunctServerStatus:
    socket_path: str

    def to_json(self) -> Dict[str, object]:
        return {"status": "defunct", "socket": self.socket_path}


@dataclasses.dataclass(frozen=True)
class AllServerStatus:
    running: List[RunningServerStatus] = dataclasses.field(default_factory=list)
    defunct: List[DefunctServerStatus] = dataclasses.field(default_factory=list)

    def to_json(self) -> List[Dict[str, Any]]:
        return [status.to_json() for status in self.running] + [
            status.to_json() for status in self.defunct
        ]


def _get_server_status(
    socket_path: Path,
    flavor: identifiers.PyreFlavor,
) -> Union[RunningServerStatus, DefunctServerStatus]:
    try:
        with connections.connect(socket_path) as (
            input_channel,
            output_channel,
        ):
            output_channel.write('["GetInfo"]\n')
            return RunningServerStatus.from_server_response(
                input_channel.readline(), flavor
            )

    except connections.ConnectionFailure:
        return DefunctServerStatus(str(socket_path))


def _print_running_server_status(running_status: Sequence[RunningServerStatus]) -> None:
    if len(running_status) == 0:
        log.stdout.write("No server is currently running.\n")
    else:
        log.stdout.write("Running Servers:\n\n")
        log.stdout.write(
            tabulate.tabulate(
                [
                    [
                        status.pid,
                        status.global_root,
                        status.relative_local_root or "",
                        status.version,
                        status.flavor,
                    ]
                    for status in running_status
                ],
                headers=[
                    "PID",
                    "Global Root",
                    "Relative Local Root",
                    "Version",
                    "Flavor",
                ],
            ),
        )
        log.stdout.write("\n")
    log.stdout.write("\n")


def _print_defunct_server_status(defunct_status: Sequence[DefunctServerStatus]) -> None:
    defunct_count = len(defunct_status)
    if defunct_count > 0:
        plurality = "" if defunct_count == 1 else "s"
        log.stdout.write(f"Found {defunct_count} defunct server{plurality} at:\n")
        for status in defunct_status:
            log.stdout.write(f" {status.socket_path}\n")
        log.stdout.write("\n")


def _print_server_status_json(server_status: AllServerStatus) -> None:
    log.stdout.write(json.dumps(server_status.to_json()))
    log.stdout.write("\n")


def _print_server_status(server_status: AllServerStatus, output_format: str) -> None:
    if output_format == command_arguments.TEXT:
        _print_running_server_status(server_status.running)
        _print_defunct_server_status(server_status.defunct)
    elif output_format == command_arguments.JSON:
        _print_server_status_json(server_status)


def _stop_server(socket_path: Path, flavor: identifiers.PyreFlavor) -> None:
    try:
        LOG.info(f"Stopping server at `{socket_path}...`")
        stop.stop_server(socket_path, flavor)
        LOG.info(f"Successfully stopped `{socket_path}.`")
    except connections.ConnectionFailure:
        LOG.info(f"Failed to connect to `{socket_path}`. Removing it...")
        stop.remove_socket_if_exists(socket_path)
    except Exception as error:
        LOG.warning(
            f"Exception occurred when trying to stop server at `{socket_path}`: {error}"
        )


def _find_server_flavor(socket_path: Path) -> identifiers.PyreFlavor:
    # Socket paths are of the form `/tmp/pyre_{md5}[__{flavor}].sock`.
    serialized_path = str(socket_path)
    for flavor in identifiers.PyreFlavor:
        if flavor.value in serialized_path:
            return flavor
    # No suffix indicates a classic server.
    return identifiers.PyreFlavor.CLASSIC


def find_all_servers(socket_paths: Iterable[Path]) -> AllServerStatus:
    running_servers = []
    defunct_servers = []

    for socket_path in socket_paths:
        flavor = _find_server_flavor(socket_path)
        status = _get_server_status(socket_path, flavor)
        if isinstance(status, RunningServerStatus):
            running_servers.append(status)
        else:
            defunct_servers.append(status)

    return AllServerStatus(running_servers, defunct_servers)


def find_all_servers_under(socket_root: Path) -> AllServerStatus:
    return find_all_servers(daemon_socket.find_socket_files(socket_root))


def run_list(output_format: str) -> commands.ExitCode:
    server_status = find_all_servers_under(daemon_socket.get_default_socket_root())
    _print_server_status(server_status, output_format)
    return commands.ExitCode.SUCCESS


def run_stop() -> commands.ExitCode:
    for socket_path in daemon_socket.find_socket_files(
        daemon_socket.get_default_socket_root()
    ):
        flavor = _find_server_flavor(socket_path)
        _stop_server(socket_path, flavor)
    LOG.info("Done\n")
    return commands.ExitCode.SUCCESS



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module defines logic for loading `.pyre_configuration` json
files into Python.

The code is broken into two parts:
- PartialConfiguration, which loads json directly
- Configuration, which is built from PartialConfiguration

This is because we allow developers (including ourselves) to
extend Pyre with a notion of sub-projects, where configuration
is split between a repository-wide .pyre_configuration and a
project-specific .pyre_configuration.local.

Out-of-the-box, open-source pyre only uses a single
.pyre_configuration because this is simpler and works for most
cases.
"""

import dataclasses
import glob
import json
import logging
import site
import sys
from dataclasses import field
from logging import Logger
from pathlib import Path
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Type,
    TypeVar,
    Union,
)

import psutil
import tomli as tomllib

from .. import command_arguments, dataclasses_merge, find_directories, identifiers
from ..filesystem import expand_global_root, expand_relative_path
from ..find_directories import (
    get_relative_local_root,
    JSON_CONFIGURATION_FILE,
    LOCAL_CONFIGURATION_FILE,
    TOML_CONFIGURATION_FILE,
)
from . import (
    exceptions,
    extension,
    platform_aware,
    python_version as python_version_module,
    search_path as search_path_module,
    shared_memory as shared_memory_module,
    site_packages,
    unwatched,
)

LOG: Logger = logging.getLogger(__name__)
T = TypeVar("T")


def _get_optional_value(source: Optional[T], default: T) -> T:
    return source if source is not None else default


def _expand_glob(pattern: str) -> List[str]:
    expanded = glob.glob(pattern)
    return [pattern] if len(expanded) == 0 else expanded


def _expand_all_globs(patterns: Iterable[str]) -> List[str]:
    return [expanded for pattern in patterns for expanded in _expand_glob(pattern)]


def get_default_site_roots() -> List[str]:
    try:
        return [site.getusersitepackages()] + site.getsitepackages()
    except AttributeError:
        # There are a few Python versions that ship with a broken venv,
        # where `getsitepackages` is not available.
        LOG.warning(
            "Either `site.getusersitepackages()` or `site.getsitepackages()` "
            + "is not available in your virtualenv. This is a known virtualenv "
            + 'bug and as a workaround please explicitly specify `"site_root"` '
            + "in your Pyre configuration."
        )
        return []


@dataclasses_merge.dataclass_merge
@dataclasses.dataclass(frozen=True)
class PartialConfiguration:
    binary: Optional[str] = None
    buck_mode: Optional[platform_aware.PlatformAware[str]] = field(
        default=None,
        metadata={"merge_policy": platform_aware.PlatformAware.merge_optional},
    )
    bxl_builder: Optional[str] = None
    only_check_paths: Sequence[str] = field(
        default_factory=list,
        metadata={"merge_policy": dataclasses_merge.Policy.PREPEND},
    )
    dot_pyre_directory: Optional[Path] = None
    enable_readonly_analysis: Optional[bool] = None
    enable_strict_override_check: Optional[bool] = None
    enable_strict_any_check: Optional[bool] = None
    enable_unawaited_awaitable_analysis: Optional[bool] = None
    excludes: Sequence[str] = field(
        default_factory=list,
        metadata={"merge_policy": dataclasses_merge.Policy.PREPEND},
    )
    extensions: Sequence[extension.Element] = field(
        default_factory=list,
        metadata={"merge_policy": dataclasses_merge.Policy.PREPEND},
    )
    ignore_all_errors: Sequence[str] = field(
        default_factory=list,
        metadata={"merge_policy": dataclasses_merge.Policy.PREPEND},
    )
    include_suppressed_errors: Optional[bool] = None
    isolation_prefix: Optional[str] = None
    logger: Optional[str] = None
    number_of_workers: Optional[int] = None
    max_number_of_workers: Optional[int] = None
    oncall: Optional[str] = None
    other_critical_files: Sequence[str] = field(
        default_factory=list,
        metadata={"merge_policy": dataclasses_merge.Policy.PREPEND},
    )
    pysa_version_hash: Optional[str] = None
    python_version: Optional[python_version_module.PythonVersion] = None
    system_platform: Optional[str] = None
    search_path: Sequence[search_path_module.RawElement] = field(
        default_factory=list,
        metadata={"merge_policy": dataclasses_merge.Policy.PREPEND},
    )
    optional_search_path: Sequence[search_path_module.RawElement] = field(
        default_factory=list,
        metadata={"merge_policy": dataclasses_merge.Policy.PREPEND},
    )
    shared_memory: shared_memory_module.SharedMemory = (
        shared_memory_module.SharedMemory()
    )
    site_package_search_strategy: Optional[site_packages.SearchStrategy] = None
    site_roots: Optional[Sequence[str]] = None
    source_directories: Optional[Sequence[search_path_module.RawElement]] = field(
        default=None,
        metadata={"merge_policy": dataclasses_merge.Policy.RAISE_WHEN_OVERWRITTEN},
    )
    strict: Optional[bool] = None
    taint_models_path: Sequence[str] = field(
        default_factory=list,
        metadata={"merge_policy": dataclasses_merge.Policy.PREPEND},
    )
    targets: Optional[Sequence[str]] = field(
        default=None,
        metadata={"merge_policy": dataclasses_merge.Policy.RAISE_WHEN_OVERWRITTEN},
    )
    typeshed: Optional[str] = None
    unwatched_dependency: Optional[unwatched.UnwatchedDependency] = None
    use_buck2: Optional[bool] = None
    version_hash: Optional[str] = None

    @staticmethod
    def _get_extra_keys() -> Set[str]:
        return {
            "allow_shadow_language_server",
            "create_open_source_configuration",
            "saved_state",
            "saved_state_metadata",
            "stable_client",
            "taint_models_path",
            "unstable_client",
            "vscode_project_identifier",
            "vscode_excluded_directories",
            "migration_status",
        }

    @staticmethod
    def from_command_arguments(
        arguments: command_arguments.CommandArguments,
    ) -> "PartialConfiguration":
        strict: Optional[bool] = True if arguments.strict else None
        source_directories = [
            search_path_module.SimpleRawElement(element)
            for element in arguments.source_directories
        ] or None
        targets: Optional[List[str]] = (
            arguments.targets if len(arguments.targets) > 0 else None
        )
        python_version_string = arguments.python_version
        return PartialConfiguration(
            binary=arguments.binary,
            buck_mode=platform_aware.PlatformAware.from_json(
                arguments.buck_mode, "buck_mode"
            ),
            only_check_paths=arguments.only_check_paths,
            dot_pyre_directory=arguments.dot_pyre_directory,
            enable_unawaited_awaitable_analysis=(
                arguments.enable_unawaited_awaitable_analysis
            ),
            excludes=arguments.exclude,
            extensions=[],
            ignore_all_errors=[],
            include_suppressed_errors=arguments.include_suppressed_errors,
            isolation_prefix=arguments.isolation_prefix,
            logger=arguments.logger,
            number_of_workers=arguments.number_of_workers,
            max_number_of_workers=arguments.max_number_of_workers,
            oncall=None,
            other_critical_files=[],
            pysa_version_hash=None,
            python_version=(
                python_version_module.PythonVersion.from_string(python_version_string)
                if python_version_string is not None
                else None
            ),
            system_platform=arguments.system_platform,
            search_path=[
                search_path_module.SimpleRawElement(element)
                for element in arguments.search_path
            ],
            optional_search_path=[
                search_path_module.SimpleRawElement(element)
                for element in arguments.optional_search_path
            ],
            shared_memory=shared_memory_module.SharedMemory(
                heap_size=arguments.shared_memory_heap_size,
                dependency_table_power=arguments.shared_memory_dependency_table_power,
                hash_table_power=arguments.shared_memory_hash_table_power,
            ),
            site_package_search_strategy=None,
            site_roots=None,
            source_directories=source_directories,
            strict=strict,
            taint_models_path=[],
            targets=targets,
            typeshed=arguments.typeshed,
            unwatched_dependency=None,
            use_buck2=arguments.use_buck2,
            version_hash=None,
        )

    @staticmethod
    def from_dict(configuration_json: Dict[str, Any]) -> "PartialConfiguration":
        def is_list_of_string(elements: object) -> bool:
            return isinstance(elements, list) and all(
                isinstance(element, str) for element in elements
            )

        def ensure_option_type(
            json: Dict[str, Any], name: str, expected_type: Type[T]
        ) -> Optional[T]:
            result = json.pop(name, None)
            if result is None:
                return None
            elif isinstance(result, expected_type):
                return result
            raise exceptions.InvalidConfiguration(
                f"Configuration field `{name}` is expected to have type "
                f"{expected_type} but got: `{result}`."
            )

        def ensure_optional_string_or_string_dict(
            json: Dict[str, Any], name: str
        ) -> Optional[Union[Dict[str, str], str]]:
            result = json.pop(name, None)
            if result is None:
                return None
            elif isinstance(result, str):
                return result
            elif isinstance(result, Dict):
                for value in result.values():
                    if not isinstance(value, str):
                        raise exceptions.InvalidConfiguration(
                            f"Configuration field `{name}` is expected to be a "
                            + f"dict of strings but got `{result}`."
                        )
                return result
            raise exceptions.InvalidConfiguration(
                f"Configuration field `{name}` is expected to be a string or a "
                + f"dict of strings but got `{result}`."
            )

        def ensure_optional_string_list(
            json: Dict[str, Any], name: str
        ) -> Optional[List[str]]:
            result = json.pop(name, None)
            if result is None:
                return None
            elif is_list_of_string(result):
                return result
            raise exceptions.InvalidConfiguration(
                f"Configuration field `{name}` is expected to be a list of "
                + f"strings but got `{result}`."
            )

        def ensure_string_list(
            json: Dict[str, Any], name: str, allow_single_string: bool = False
        ) -> List[str]:
            result = json.pop(name, [])
            if allow_single_string and isinstance(result, str):
                result = [result]
            if is_list_of_string(result):
                return result
            raise exceptions.InvalidConfiguration(
                f"Configuration field `{name}` is expected to be a list of "
                + f"strings but got `{result}`."
            )

        def ensure_list(json: Dict[str, object], name: str) -> List[object]:
            result = json.pop(name, [])
            if isinstance(result, list):
                return result
            raise exceptions.InvalidConfiguration(
                f"Configuration field `{name}` is expected to be a list but got `{result}`."
            )

        def create_search_paths(
            search_path_json: Union[Dict[str, object], str],
        ) -> List[search_path_module.RawElement]:
            if isinstance(search_path_json, list):
                search_path = [
                    search_path_module.create_raw_element(json)
                    for json in search_path_json
                ]
            else:
                search_path = [search_path_module.create_raw_element(search_path_json)]
            return search_path

        try:
            dot_pyre_directory = ensure_option_type(
                configuration_json, "dot_pyre_directory", str
            )

            search_path, optional_search_path = (
                create_search_paths(configuration_json.pop(json_name, []))
                for json_name in ("search_path", "optional_search_path")
            )

            python_version_json = configuration_json.pop("python_version", None)
            if python_version_json is None:
                python_version = None
            elif isinstance(python_version_json, str):
                python_version = python_version_module.PythonVersion.from_string(
                    python_version_json
                )
            else:
                raise exceptions.InvalidConfiguration(
                    "Expect python version to be a string but got"
                    + f"'{python_version_json}'"
                )

            shared_memory_json = ensure_option_type(
                configuration_json, "shared_memory", dict
            )
            if shared_memory_json is None:
                shared_memory = shared_memory_module.SharedMemory()
            else:
                shared_memory = shared_memory_module.SharedMemory(
                    heap_size=ensure_option_type(shared_memory_json, "heap_size", int),
                    dependency_table_power=ensure_option_type(
                        shared_memory_json, "dependency_table_power", int
                    ),
                    hash_table_power=ensure_option_type(
                        shared_memory_json, "hash_table_power", int
                    ),
                )
                for unrecognized_key in shared_memory_json:
                    LOG.warning(f"Unrecognized configuration item: {unrecognized_key}")

            source_directories_json = ensure_option_type(
                configuration_json, "source_directories", list
            )
            if isinstance(source_directories_json, list):
                source_directories = [
                    search_path_module.create_raw_element(json)
                    for json in source_directories_json
                ]
            else:
                source_directories = None

            site_package_search_strategy_json = ensure_option_type(
                configuration_json, "site_package_search_strategy", str
            )
            if site_package_search_strategy_json is None:
                site_package_search_strategy = None
            else:
                site_package_search_strategy = site_packages.SearchStrategy.from_string(
                    site_package_search_strategy_json
                )
                if site_package_search_strategy is None:
                    raise exceptions.InvalidConfiguration(
                        "Invalid value for `site_package_search_strategy`: "
                        f"{site_package_search_strategy_json}. Available choices: "
                        f"{[str(x) for x in site_packages.SearchStrategy]}."
                    )

            unwatched_dependency_json = ensure_option_type(
                configuration_json, "unwatched_dependency", dict
            )
            if unwatched_dependency_json is None:
                unwatched_dependency = None
            else:
                unwatched_dependency = unwatched.UnwatchedDependency.from_json(
                    unwatched_dependency_json
                )

            partial_configuration = PartialConfiguration(
                binary=ensure_option_type(configuration_json, "binary", str),
                buck_mode=platform_aware.PlatformAware.from_json(
                    ensure_optional_string_or_string_dict(
                        configuration_json, "buck_mode"
                    ),
                    "buck_mode",
                ),
                bxl_builder=ensure_option_type(configuration_json, "bxl_builder", str),
                only_check_paths=ensure_string_list(
                    configuration_json, "only_check_paths"
                ),
                dot_pyre_directory=(
                    Path(dot_pyre_directory) if dot_pyre_directory is not None else None
                ),
                enable_readonly_analysis=ensure_option_type(
                    configuration_json, "enable_readonly_analysis", bool
                ),
                enable_strict_override_check=ensure_option_type(
                    configuration_json, "enable_strict_override_check", bool
                ),
                enable_strict_any_check=ensure_option_type(
                    configuration_json, "enable_strict_any_check", bool
                ),
                enable_unawaited_awaitable_analysis=ensure_option_type(
                    configuration_json, "enable_unawaited_awaitable_analysis", bool
                ),
                excludes=ensure_string_list(
                    configuration_json, "exclude", allow_single_string=True
                ),
                extensions=[
                    extension.Element.from_json(json)
                    for json in ensure_list(configuration_json, "extensions")
                ],
                ignore_all_errors=ensure_string_list(
                    configuration_json, "ignore_all_errors"
                ),
                include_suppressed_errors=ensure_option_type(
                    configuration_json, "include_suppressed_errors", bool
                ),
                isolation_prefix=ensure_option_type(
                    configuration_json, "isolation_prefix", str
                ),
                logger=ensure_option_type(configuration_json, "logger", str),
                number_of_workers=ensure_option_type(
                    configuration_json, "workers", int
                ),
                max_number_of_workers=ensure_option_type(
                    configuration_json, "max_workers", int
                ),
                oncall=ensure_option_type(configuration_json, "oncall", str),
                other_critical_files=ensure_string_list(
                    configuration_json, "critical_files"
                ),
                pysa_version_hash=ensure_option_type(
                    configuration_json, "pysa_version", str
                ),
                python_version=python_version,
                system_platform=ensure_option_type(
                    configuration_json, "system_platform", str
                ),
                search_path=search_path,
                optional_search_path=optional_search_path,
                shared_memory=shared_memory,
                site_package_search_strategy=site_package_search_strategy,
                site_roots=ensure_optional_string_list(
                    configuration_json, "site_roots"
                ),
                source_directories=source_directories,
                strict=ensure_option_type(configuration_json, "strict", bool),
                taint_models_path=ensure_string_list(
                    configuration_json, "taint_models_path", allow_single_string=True
                ),
                targets=ensure_optional_string_list(configuration_json, "targets"),
                typeshed=ensure_option_type(configuration_json, "typeshed", str),
                unwatched_dependency=unwatched_dependency,
                use_buck2=ensure_option_type(configuration_json, "use_buck2", bool),
                version_hash=ensure_option_type(configuration_json, "version", str),
            )

            # Check for unused keys
            extra_keys = PartialConfiguration._get_extra_keys()
            for unrecognized_key in configuration_json:
                if unrecognized_key not in extra_keys:
                    LOG.warning(f"Unrecognized configuration item: {unrecognized_key}")

            return partial_configuration
        except json.JSONDecodeError as error:
            raise exceptions.InvalidConfiguration(
                "Invalid configuration file"
            ) from error

    @staticmethod
    def from_file(path: Path) -> "PartialConfiguration":
        is_toml = path.suffix.lower() == ".toml"
        try:
            contents = path.read_text(encoding="utf-8")
            if is_toml:
                configuration_toml = tomllib.loads(contents)
                configuration_tool = (
                    configuration_toml["tool"] if "tool" in configuration_toml else {}
                )
                configuration_pyre = (
                    configuration_tool["pyre"] if "pyre" in configuration_tool else {}
                )
                return PartialConfiguration.from_dict(configuration_pyre)
            else:
                configuration_json = json.loads(contents)
                return PartialConfiguration.from_dict(configuration_json)
        except OSError as error:
            raise exceptions.InvalidConfiguration(
                f"Error when reading {path}"
            ) from error
        except json.JSONDecodeError as error:
            raise exceptions.InvalidConfiguration(
                f"Invalid JSON file at {path}"
            ) from error
        except tomllib.TOMLDecodeError as error:
            raise exceptions.InvalidConfiguration(
                f"Invalid TOML file at {path}"
            ) from error

    def expand_relative_paths(self, root: str) -> "PartialConfiguration":
        unwatched_dependency = self.unwatched_dependency
        if unwatched_dependency is not None:
            files = unwatched_dependency.files
            unwatched_dependency = unwatched.UnwatchedDependency(
                change_indicator=unwatched_dependency.change_indicator,
                files=unwatched.UnwatchedFiles(
                    root=expand_relative_path(root, files.root),
                    checksum_path=files.checksum_path,
                ),
            )
        return dataclasses.replace(
            self,
            binary=(
                expand_relative_path(root, self.binary)
                if self.binary is not None
                else self.binary
            ),
            only_check_paths=[
                expand_relative_path(root, path) for path in self.only_check_paths
            ],
            ignore_all_errors=[
                expand_relative_path(root, path) for path in self.ignore_all_errors
            ],
            logger=(
                expand_relative_path(root, self.logger)
                if self.logger is not None
                else self.logger
            ),
            other_critical_files=[
                expand_relative_path(root, path) for path in self.other_critical_files
            ],
            search_path=[path.expand_relative_root(root) for path in self.search_path],
            optional_search_path=[
                path.expand_relative_root(root) for path in self.optional_search_path
            ],
            site_package_search_strategy=self.site_package_search_strategy,
            source_directories=(
                [path.expand_relative_root(root) for path in self.source_directories]
                if self.source_directories is not None
                else self.source_directories
            ),
            taint_models_path=[
                expand_relative_path(root, path) for path in self.taint_models_path
            ],
            typeshed=(
                expand_relative_path(root, self.typeshed)
                if self.typeshed is not None
                else self.typeshed
            ),
            unwatched_dependency=unwatched_dependency,
        )


def merge_partial_configurations(
    base: PartialConfiguration, override: PartialConfiguration
) -> PartialConfiguration:
    try:
        # pyre-ignore[16]: Pyre does not understand `dataclass_merge`
        return PartialConfiguration.merge(base, override)
    except dataclasses_merge.DataclassMergeError as error:
        raise exceptions.InvalidConfiguration(str(error)) from None


@dataclasses.dataclass(frozen=True)
class Configuration:
    global_root: Path

    binary: Optional[str] = None
    buck_mode: Optional[platform_aware.PlatformAware[str]] = None
    bxl_builder: Optional[str] = None
    only_check_paths: Sequence[str] = field(default_factory=list)
    dot_pyre_directory: Optional[Path] = None
    enable_readonly_analysis: Optional[bool] = None
    enable_strict_override_check: Optional[bool] = None
    enable_strict_any_check: Optional[bool] = None
    enable_unawaited_awaitable_analysis: Optional[bool] = None
    excludes: Sequence[str] = field(default_factory=list)
    extensions: Sequence[extension.Element] = field(default_factory=list)
    ignore_all_errors: Sequence[str] = field(default_factory=list)
    include_suppressed_errors: Optional[bool] = None
    isolation_prefix: Optional[str] = None
    logger: Optional[str] = None
    number_of_workers: Optional[int] = None
    max_number_of_workers: Optional[int] = None
    oncall: Optional[str] = None
    other_critical_files: Sequence[str] = field(default_factory=list)
    pysa_version_hash: Optional[str] = None
    python_version: Optional[python_version_module.PythonVersion] = None
    system_platform: Optional[str] = None
    relative_local_root: Optional[str] = None
    search_path: Sequence[search_path_module.RawElement] = field(default_factory=list)
    optional_search_path: Sequence[search_path_module.RawElement] = field(
        default_factory=list
    )
    shared_memory: shared_memory_module.SharedMemory = (
        shared_memory_module.SharedMemory()
    )
    site_package_search_strategy: site_packages.SearchStrategy = (
        site_packages.SearchStrategy.NONE
    )
    site_roots: Optional[Sequence[str]] = None
    source_directories: Optional[Sequence[search_path_module.RawElement]] = None
    strict: bool = False
    taint_models_path: Sequence[str] = field(default_factory=list)
    targets: Optional[Sequence[str]] = None
    typeshed: Optional[str] = None
    unwatched_dependency: Optional[unwatched.UnwatchedDependency] = None
    use_buck2: bool = True
    version_hash: Optional[str] = None

    @staticmethod
    def from_partial_configuration(
        global_root: Path,
        relative_local_root: Optional[str],
        partial_configuration: PartialConfiguration,
    ) -> "Configuration":
        search_path = partial_configuration.search_path
        optional_search_path = partial_configuration.optional_search_path
        ignore_all_errors = partial_configuration.ignore_all_errors
        only_check_paths = partial_configuration.only_check_paths

        return Configuration(
            global_root=global_root,
            dot_pyre_directory=partial_configuration.dot_pyre_directory,
            binary=partial_configuration.binary,
            buck_mode=partial_configuration.buck_mode,
            bxl_builder=partial_configuration.bxl_builder,
            only_check_paths=[
                expand_global_root(path, global_root=str(global_root))
                for path in only_check_paths
            ],
            enable_readonly_analysis=partial_configuration.enable_readonly_analysis,
            enable_strict_override_check=partial_configuration.enable_strict_override_check,
            enable_strict_any_check=partial_configuration.enable_strict_any_check,
            enable_unawaited_awaitable_analysis=(
                partial_configuration.enable_unawaited_awaitable_analysis
            ),
            excludes=partial_configuration.excludes,
            extensions=partial_configuration.extensions,
            ignore_all_errors=_expand_all_globs(
                expand_global_root(path, global_root=str(global_root))
                for path in ignore_all_errors
            ),
            include_suppressed_errors=partial_configuration.include_suppressed_errors,
            isolation_prefix=partial_configuration.isolation_prefix,
            logger=partial_configuration.logger,
            number_of_workers=partial_configuration.number_of_workers,
            max_number_of_workers=partial_configuration.max_number_of_workers,
            oncall=partial_configuration.oncall,
            other_critical_files=partial_configuration.other_critical_files,
            pysa_version_hash=partial_configuration.pysa_version_hash,
            python_version=partial_configuration.python_version,
            system_platform=partial_configuration.system_platform,
            relative_local_root=relative_local_root,
            search_path=[
                path.expand_global_root(str(global_root)) for path in search_path
            ],
            optional_search_path=[
                path.expand_global_root(str(global_root))
                for path in optional_search_path
            ],
            shared_memory=partial_configuration.shared_memory,
            site_package_search_strategy=partial_configuration.site_package_search_strategy
            or site_packages.SearchStrategy.NONE,
            site_roots=partial_configuration.site_roots,
            source_directories=partial_configuration.source_directories,
            strict=_get_optional_value(partial_configuration.strict, default=False),
            taint_models_path=partial_configuration.taint_models_path,
            targets=partial_configuration.targets,
            typeshed=partial_configuration.typeshed,
            unwatched_dependency=partial_configuration.unwatched_dependency,
            use_buck2=_get_optional_value(
                partial_configuration.use_buck2, default=True
            ),
            version_hash=partial_configuration.version_hash,
        )

    @property
    def project_identifier(self) -> str:
        """
        Note: it is important that this identifier, which is part of what determines
        the socket path for connecting to an ocaml daemon, is entirely determined based
        on fields that come from the command arguments.
        """
        return identifiers.get_project_identifier(
            self.global_root,
            self.relative_local_root,
        )

    def to_json(self) -> Dict[str, object]:
        """
        This method is for display purpose only. Do *NOT* expect this method
        to produce JSONs that can be de-serialized back into configurations.
        """
        binary = self.binary
        buck_mode = self.buck_mode
        bxl_builder = self.bxl_builder
        isolation_prefix = self.isolation_prefix
        logger = self.logger
        number_of_workers = self.number_of_workers
        max_number_of_workers = self.max_number_of_workers
        oncall = self.oncall
        pysa_version_hash = self.pysa_version_hash
        python_version = self.python_version
        relative_local_root = self.relative_local_root
        source_directories = self.source_directories
        site_package_search_strategy = self.site_package_search_strategy
        site_roots = self.site_roots
        targets = self.targets
        typeshed = self.typeshed
        unwatched_dependency = self.unwatched_dependency
        version_hash = self.version_hash
        return {
            "global_root": str(self.global_root),
            "dot_pyre_directory": str(self.dot_pyre_directory),
            **({"binary": binary} if binary is not None else {}),
            **({"buck_mode": buck_mode.to_json()} if buck_mode is not None else {}),
            **({"bxl_builder": bxl_builder} if bxl_builder is not None else {}),
            "only_check_paths": list(self.only_check_paths),
            **(
                {"enable_readonly_analysis": self.enable_readonly_analysis}
                if self.enable_readonly_analysis is not None
                else {}
            ),
            **(
                {"enable_strict_override_check": self.enable_strict_override_check}
                if self.enable_strict_override_check is not None
                else {}
            ),
            **(
                {"enable_strict_any_check": self.enable_strict_any_check}
                if self.enable_strict_any_check is not None
                else {}
            ),
            **(
                {
                    "enable_unawaited_awaitable_analysis": (
                        self.enable_unawaited_awaitable_analysis
                    )
                }
                if self.enable_unawaited_awaitable_analysis is not None
                else {}
            ),
            "excludes": list(self.excludes),
            "extensions": [extension.to_json() for extension in self.extensions],
            "ignore_all_errors": list(self.ignore_all_errors),
            **(
                {"include_suppressed_errors": self.include_suppressed_errors}
                if self.include_suppressed_errors is not None
                else {}
            ),
            **(
                {"isolation_prefix": isolation_prefix}
                if isolation_prefix is not None
                else {}
            ),
            **({"logger": logger} if logger is not None else {}),
            **({"oncall": oncall} if oncall is not None else {}),
            **({"workers": number_of_workers} if number_of_workers is not None else {}),
            **(
                {"max_workers": max_number_of_workers}
                if max_number_of_workers is not None
                else {}
            ),
            "other_critical_files": list(self.other_critical_files),
            **(
                {"pysa_version_hash": pysa_version_hash}
                if pysa_version_hash is not None
                else {}
            ),
            **(
                {"python_version": python_version.to_string()}
                if python_version is not None
                else {}
            ),
            **(
                {"system_platform": self.system_platform}
                if self.system_platform is not None
                else {}
            ),
            **(
                {"relative_local_root": relative_local_root}
                if relative_local_root is not None
                else {}
            ),
            "search_path": [str(path) for path in self.search_path],
            "optional_search_path": [str(path) for path in self.optional_search_path],
            **(
                {"shared_memory": self.shared_memory.to_json()}
                if self.shared_memory != shared_memory_module.SharedMemory()
                else {}
            ),
            **(
                {"site_package_search_strategy": site_package_search_strategy}
                if site_package_search_strategy is not None
                else {}
            ),
            "site_roots": site_roots if site_roots is not None else [],
            **(
                {"source_directories": [str(path) for path in source_directories]}
                if source_directories is not None
                else {}
            ),
            "strict": self.strict,
            "taint_models_path": list(self.taint_models_path),
            **({"targets": list(targets)} if targets is not None else {}),
            **({"typeshed": typeshed} if typeshed is not None else {}),
            **(
                {"unwatched_dependency": unwatched_dependency.to_json()}
                if unwatched_dependency is not None
                else {}
            ),
            "use_buck2": self.use_buck2,
            **({"version_hash": version_hash} if version_hash is not None else {}),
        }

    def get_existent_unwatched_dependency(
        self,
    ) -> Optional[unwatched.UnwatchedDependency]:
        unwatched_dependency = self.unwatched_dependency
        if unwatched_dependency is None:
            return None
        unwatched_root = Path(unwatched_dependency.files.root)
        try:
            if not unwatched_root.is_dir():
                LOG.warning(
                    "Nonexistent directory passed in to `unwatched_dependency`: "
                    f"`{unwatched_root}`"
                )
                return None
            checksum_path = unwatched_root / unwatched_dependency.files.checksum_path
            if not checksum_path.is_file():
                LOG.warning(
                    "Nonexistent file passed in to `unwatched_dependency`: "
                    f"`{checksum_path}`"
                )
                return None
            return self.unwatched_dependency
        except PermissionError as error:
            LOG.warning(str(error))
            return None

    def get_site_roots(self) -> Sequence[str]:
        site_roots = self.site_roots
        if site_roots is not None:
            return site_roots
        return get_default_site_roots()

    def expand_and_get_existent_search_paths(
        self,
    ) -> List[search_path_module.Element]:
        site_roots = self.get_site_roots()
        existent_paths = search_path_module.process_raw_elements(
            self.search_path, site_roots, required=True
        ) + search_path_module.process_raw_elements(
            self.optional_search_path, site_roots
        )

        site_packages_paths = site_packages.search_for_paths(
            self.site_package_search_strategy, site_roots
        )

        return existent_paths + site_packages_paths

    def expand_and_get_existent_source_directories(
        self,
    ) -> List[search_path_module.Element]:
        source_directories = self.source_directories
        if source_directories is not None:
            return search_path_module.process_raw_elements(
                source_directories, self.get_site_roots()
            )
        else:
            return []

    def get_number_of_workers(self) -> int:
        number_of_workers = self.number_of_workers
        if number_of_workers is not None and number_of_workers > 0:
            return number_of_workers

        number_of_logical_cores = psutil.cpu_count(logical=True)
        if number_of_logical_cores is None:
            default_number_of_workers = 1
        else:
            default_number_of_workers = max(1, number_of_logical_cores // 2 - 1)
            max_number_of_workers = self.max_number_of_workers
            if (
                max_number_of_workers is not None
                and max_number_of_workers > 0
                and default_number_of_workers > max_number_of_workers
            ):
                LOG.info(
                    f"The number of workers is capped at the maximum {max_number_of_workers}."
                )
                default_number_of_workers = max_number_of_workers

        LOG.info(
            "Could not determine the number of Pyre workers from configuration. "
            f"Auto-set the value to {default_number_of_workers}."
        )
        if default_number_of_workers <= 1:
            LOG.info(
                "Consider setting the `--sequential` flag instead when the number "
                "of parallel workers is not greater than 1."
            )
        return default_number_of_workers

    def get_valid_extension_suffixes(self) -> List[str]:
        vaild_extensions = []
        for element in self.extensions:
            if not element.suffix.startswith("."):
                LOG.warning(
                    "Filtering out extension which does not start with `.`: "
                    f"`{element.suffix}`"
                )
            else:
                vaild_extensions.append(element.command_line_argument())
        return vaild_extensions

    def get_python_version(self) -> python_version_module.PythonVersion:
        python_version = self.python_version
        if python_version is not None:
            return python_version
        else:
            version_info = sys.version_info
            return python_version_module.PythonVersion(
                major=version_info.major,
                minor=version_info.minor,
                micro=version_info.micro,
            )

    def get_system_platform(self) -> str:
        system_platform = self.system_platform
        if system_platform is not None:
            return system_platform
        else:
            return sys.platform


def create_configuration(
    arguments: command_arguments.CommandArguments, base_directory: Path
) -> Configuration:
    local_root_argument = arguments.local_configuration
    search_base = (
        base_directory
        if local_root_argument is None
        else base_directory / local_root_argument
    )
    found_root = find_directories.find_global_and_local_root(search_base)

    # If the local root was explicitly specified but does not exist, return an
    # error instead of falling back to current directory.
    if local_root_argument is not None:
        if found_root is None:
            raise exceptions.InvalidConfiguration(
                "A local configuration path was explicitly specified, but neither"
                + f" {JSON_CONFIGURATION_FILE} nor {TOML_CONFIGURATION_FILE} file was found in {search_base}"
                + " or its parents."
            )
        elif found_root.local_root is None:
            raise exceptions.InvalidConfiguration(
                "A local configuration path was explicitly specified, but no"
                + f" {LOCAL_CONFIGURATION_FILE} file was found in {search_base}"
                + " or its parents."
            )

    command_argument_configuration = PartialConfiguration.from_command_arguments(
        arguments
    ).expand_relative_paths(str(Path.cwd()))
    if found_root is None:
        project_root = Path.cwd()
        relative_local_root = None
        partial_configuration = command_argument_configuration
    else:
        project_root = found_root.global_root
        relative_local_root = None
        if (project_root / JSON_CONFIGURATION_FILE).is_file():
            partial_configuration = PartialConfiguration.from_file(
                project_root / JSON_CONFIGURATION_FILE
            ).expand_relative_paths(str(project_root))
        else:
            LOG.debug(
                "Could not find `.pyre_configuration` in the project root.Searching for `pyproject.toml`..."
            )
            partial_configuration = PartialConfiguration.from_file(
                project_root / TOML_CONFIGURATION_FILE
            ).expand_relative_paths(str(project_root))
        local_root = found_root.local_root
        if local_root is not None:
            relative_local_root = get_relative_local_root(project_root, local_root)
            partial_configuration = merge_partial_configurations(
                base=partial_configuration,
                override=PartialConfiguration.from_file(
                    local_root / LOCAL_CONFIGURATION_FILE
                ).expand_relative_paths(str(local_root)),
            )
        partial_configuration = merge_partial_configurations(
            base=partial_configuration,
            override=command_argument_configuration,
        )
        if arguments.no_logger:
            partial_configuration = dataclasses.replace(
                partial_configuration, logger=None
            )

    return Configuration.from_partial_configuration(
        project_root, relative_local_root, partial_configuration
    )


def create_overridden_configuration(
    arguments: command_arguments.CommandArguments,
    base_directory: Path,
    configuration: str,
) -> Configuration:
    if arguments.local_configuration:
        LOG.warning(
            f"Local configuration provided but skipped due to overridden global configuration {base_directory / configuration}"
        )
    command_argument_configuration = PartialConfiguration.from_command_arguments(
        arguments
    ).expand_relative_paths(str(base_directory))

    partial_configuration = merge_partial_configurations(
        base=PartialConfiguration.from_file(
            base_directory / configuration
        ).expand_relative_paths(str(base_directory)),
        override=command_argument_configuration,
    )
    if arguments.no_logger:
        partial_configuration = dataclasses.replace(partial_configuration, logger=None)
    return Configuration.from_partial_configuration(
        base_directory, None, partial_configuration
    )


def check_nested_local_configuration(configuration: Configuration) -> None:
    """
    Raises `InvalidConfiguration` if the check fails.
    """
    relative_local_root = configuration.relative_local_root
    if relative_local_root is None:
        return

    def is_subdirectory(child: Path, parent: Path) -> bool:
        return parent == child or parent in child.parents

    # We search from the parent of the local root, looking for another local
    # configuration file that lives above the current one
    local_root_path = (configuration.global_root / relative_local_root).resolve()
    current_directory = local_root_path.parent
    while True:
        found_root = find_directories.find_global_and_local_root(current_directory)
        if found_root is None:
            break

        nesting_local_root = found_root.local_root
        if nesting_local_root is None:
            break

        nesting_configuration = PartialConfiguration.from_file(
            nesting_local_root / LOCAL_CONFIGURATION_FILE
        ).expand_relative_paths(str(nesting_local_root))
        nesting_ignored_all_errors_path = _expand_all_globs(
            expand_global_root(path, global_root=str(found_root.global_root))
            for path in nesting_configuration.ignore_all_errors
        )
        if not any(
            is_subdirectory(child=local_root_path, parent=Path(path))
            for path in nesting_ignored_all_errors_path
        ):
            error_message = (
                "Local configuration is nested under another local configuration at "
                f"`{nesting_local_root}`.\nPlease add `{local_root_path}` to the "
                "`ignore_all_errors` field of the parent, or combine the sources "
                "into a single configuration, or split the parent configuration to "
                "avoid inconsistent errors."
            )
            raise exceptions.InvalidConfiguration(error_message)
        current_directory = nesting_local_root.parent



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
This module provides the logic for the `pyre rage` command, which collects
information that may be useful for debugging problems with Pyre (e.g. logs
from the client and daemon, recent repository history) and prints it to
standard output so that it can be uploaded in a bug report.
"""

import dataclasses
import datetime
import itertools
import logging
import shutil
import subprocess
from pathlib import Path
from typing import List, Optional, Sequence, TextIO, Tuple

from .. import command_arguments, frontend_configuration, identifiers, log, version
from . import commands, start

LOG: logging.Logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class Section:
    name: str
    content: str


def _print_section(section: Section, output: TextIO) -> None:
    print(f"{section.name}", file=output)
    separator = "=" * len(section.name)
    print(f"{separator}", file=output)
    print(section.content, file=output)
    print("", file=output, flush=True)


def _version_section(configuration: frontend_configuration.Base) -> Section:
    client_version_line = f"Client version: {version.__version__}"
    try:
        binary_version = configuration.get_binary_version()
        binary_version_line = f"Binary version: {binary_version}"
    except Exception as error:
        binary_version_line = f"Could not determine binary version: {error}"
    return Section(
        name="Versions", content="\n".join([client_version_line, binary_version_line])
    )


def _configuration_section(
    configuration: frontend_configuration.Base,
) -> Section:
    return Section(
        name="Configuration", content=configuration.get_content_for_display()
    )


def _get_subprocess_stdout(command: Sequence[str]) -> Optional[str]:
    # lint-ignore: NoUnsafeExecRule
    result = subprocess.run(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True
    )
    if result.returncode != 0:
        return None
    return result.stdout


def _get_file_content(path: Path) -> Optional[str]:
    try:
        return path.read_text()
    except Exception:
        return None


def _mercurial_section(
    mercurial: str, name: str, additional_flags: Optional[Sequence[str]] = None
) -> Optional[Section]:
    output = _get_subprocess_stdout(
        [mercurial, name] + ([] if additional_flags is None else list(additional_flags))
    )
    return (
        None
        if output is None
        else Section(name=f"Mercurial {name.capitalize()}", content=output)
    )


def _watchman_section(watchman: str, name: str) -> Optional[Section]:
    output = _get_subprocess_stdout([watchman, name])
    return (
        None
        if output is None
        else Section(name=f"Watchman {name.capitalize()}", content=output)
    )


def _get_server_log_timestamp_and_paths(
    log_directory: Path,
    flavor: identifiers.PyreFlavor,
) -> List[Tuple[datetime.datetime, Path]]:
    try:
        return sorted(
            (
                (timestamp, path)
                for timestamp, path in (
                    (start.datetime_from_log_path(path), path)
                    for path in (
                        log_directory / flavor.server_log_subdirectory()
                    ).iterdir()
                    if path.is_file()
                )
                if timestamp is not None
            ),
            key=lambda pair: pair[0],
            reverse=True,
        )
    except Exception:
        return []


def _server_log_sections(
    log_directory: Path, flavor: identifiers.PyreFlavor, limit: Optional[int] = None
) -> List[Section]:
    # Log files are sorted according to server start time: recently started servers
    # will come first.
    timestamp_and_paths = _get_server_log_timestamp_and_paths(
        log_directory,
        flavor,
    )

    sections: List[Section] = []
    for timestamp, path in timestamp_and_paths:
        if limit is not None and len(sections) >= limit:
            break
        content = _get_file_content(path)
        if content is None:
            continue
        sections.append(
            Section(name=f"Server Log ({flavor.value}) ({timestamp})", content=content)
        )
    return sections


def _client_log_section(log_directory: Path) -> Optional[Section]:
    content = _get_file_content(log_directory / "pyre.stderr")
    if content is None:
        return None
    return Section(name="Client Log", content=content)


def _print_configuration_sections(
    configuration: frontend_configuration.Base, output: TextIO
) -> None:
    LOG.info("Collecting information about Pyre configurations...")
    _print_section(_version_section(configuration), output)
    _print_section(_configuration_section(configuration), output)


def _print_mercurial_sections(output: TextIO) -> None:
    LOG.info("Collecting information about mercurial...")
    mercurial = shutil.which("hg")
    if mercurial is not None:
        for section in [
            _mercurial_section(mercurial, "id"),
            _mercurial_section(mercurial, "status"),
            _mercurial_section(mercurial, "diff"),
            _mercurial_section(
                mercurial, "reflog", additional_flags=["--limit", "100"]
            ),
        ]:
            if section is not None:
                _print_section(section, output)


def _print_watchman_sections(output: TextIO) -> None:
    LOG.info("Collecting information about watchman...")
    watchman = shutil.which("watchman")
    if watchman is not None:
        for section in [_watchman_section(watchman, "watch-list")]:
            if section is not None:
                _print_section(section, output)


def _print_log_file_sections(
    log_directory: Path, server_log_count: Optional[int], output: TextIO
) -> None:
    flavor = identifiers.PyreFlavor.CLASSIC
    LOG.info(
        f"Collecting information from Pyre's log files for the {flavor.value} flavor..."
    )
    for section in itertools.chain(
        _server_log_sections(log_directory, flavor, limit=server_log_count),
        [
            _client_log_section(log_directory),
        ],
    ):
        if section is not None:
            _print_section(section, output)


def run_rage(
    configuration: frontend_configuration.Base,
    arguments: command_arguments.RageArguments,
    output: TextIO,
) -> None:
    _print_configuration_sections(configuration, output)
    _print_mercurial_sections(output)
    _print_watchman_sections(output)
    _print_log_file_sections(
        configuration.get_log_directory(), arguments.server_log_count, output
    )
    LOG.info("Done\n")


def run(
    configuration: frontend_configuration.Base,
    arguments: command_arguments.RageArguments,
) -> commands.ExitCode:
    output_path = arguments.output
    if output_path is None:
        run_rage(configuration, arguments, log.stdout)
    else:
        # lint-ignore: NoUnsafeFilesystemRule
        with open(output_path) as output:
            run_rage(configuration, arguments, output)
    return commands.ExitCode.SUCCESS



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

error_code_to_fragment: dict[int, str] = {
    0: "0-unused-ignore",
    2: "2-missing-parameter-annotation",
    3: "3-missing-return-annotation",
    4: "4-missing-attribute-annotation",
    5: "5-missing-global-annotation",
    6: "6-incompatible-parameter-type",
    7: "7-incompatible-return-type",
    8: "8-incompatible-attribute-type",
    9: "9-incompatible-variable-type",
    10: "10-unbound-name",
    11: "1131-undefined-or-invalid-type",
    12: "12-incompatible-awaitable-type",
    13: "13-uninitialized-attribute",
    14: "1415-behavioral-subtyping",
    15: "1415-behavioral-subtyping",
    16: "16-missing-attributes",
    17: "17-incompatible-constructor-annotation",
    18: "1821-undefined-name-undefined-import",
    21: "1821-undefined-name-undefined-import",
    19: "19-too-many-argument",
    20: "20-missing-argument",
    22: "22-redundant-cast",
    23: "23-unable-to-unpack",
    24: "24-invalid-type-parameters",
    26: "26-typed-dictionary-access-with-non-literal",
    27: "27-typed-dictionary-key-not-found",
    28: "28-unexpected-keyword",
    29: "29-call-error",
    30: "3036-terminating-analysis,-mutually-recursive-type-variables",
    36: "3036-terminating-analysis,-mutually-recursive-type-variables",
    31: "31-invalid-type",
    32: "32-invalid-argument",
    33: "33-prohibited-any",
    34: "34-invalid-type-variable",
    35: "35-illegal-annotation-target",
    39: "39-invalid-inheritance",
    40: "40-invalid-override",
    41: "41-invalid-assignment",
    42: "42-missing-overload-implementation",
    43: "43-incompatible-overload-implementation",
    45: "45-invalid-class-instantiation",
    46: "46-invalid-type-variance",
    47: "47-invalid-method-signature",
    48: "48-invalid-exception",
    49: "49-unsafe-cast",
    51: "51-unused-local-mode",
    52: "52-private-protocol-property",
    53: "53-missing-annotation-for-captured-variables",
    54: "54-invalid-typeddict-operation",
    55: "55-typeddict-initialization-error",
    56: "56-invalid-decoration",
    57: "57-incompatible-async-generator-return-type",
    58: "58-unsupported-operand",
    59: "59-duplicate-type-variables",
    60: "60-unable-to-concatenate-tuple",
    61: "61-uninitialized-local",
    62: "62-non-literal-string",
}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

"""
TODO(T132414938) Add a module-level docstring
"""

import argparse
import logging
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from ..configuration import Configuration
from ..errors import Errors, PartialErrorSuppression
from ..filesystem import add_local_mode, LocalMode
from ..repository import Repository

LOG: logging.Logger = logging.getLogger(__name__)


class ErrorSource(Enum):
    STDIN = "stdin"
    GENERATE = "generate"

    def __repr__(self) -> str:
        return self.value


@dataclass(frozen=True)
class CommandArguments:
    comment: Optional[str]
    max_line_length: Optional[int]
    truncate: bool
    unsafe: bool
    force_format_unsuppressed: bool
    lint: bool
    no_commit: bool
    should_clean: bool

    @staticmethod
    def from_arguments(arguments: argparse.Namespace) -> "CommandArguments":
        return CommandArguments(
            comment=arguments.comment,
            max_line_length=arguments.max_line_length,
            truncate=arguments.truncate,
            unsafe=getattr(arguments, "unsafe", False),
            force_format_unsuppressed=getattr(
                arguments, "force_format_unsuppressed", False
            ),
            lint=arguments.lint,
            no_commit=arguments.no_commit,
            should_clean=arguments.should_clean,
        )


class Command:
    def __init__(self, repository: Repository) -> None:
        self._repository: Repository = repository

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        pass

    def run(self) -> None:
        pass


class ErrorSuppressingCommand(Command):
    def __init__(
        self, command_arguments: CommandArguments, repository: Repository
    ) -> None:
        super().__init__(repository)
        self._command_arguments: CommandArguments = command_arguments
        self._comment: Optional[str] = command_arguments.comment
        self._max_line_length: Optional[int] = command_arguments.max_line_length
        self._truncate: bool = command_arguments.truncate
        self._unsafe: bool = command_arguments.unsafe
        self._force_format_unsuppressed: bool = (
            command_arguments.force_format_unsuppressed
        )
        self._lint: bool = command_arguments.lint
        self._no_commit: bool = command_arguments.no_commit
        self._should_clean: bool = command_arguments.should_clean

    @classmethod
    def add_arguments(cls, parser: argparse.ArgumentParser) -> None:
        super(ErrorSuppressingCommand, ErrorSuppressingCommand).add_arguments(parser)
        parser.add_argument("--comment", help="Custom comment after fixme comments")
        parser.add_argument(
            "--max-line-length",
            default=88,
            type=int,
            help="Enforce maximum line length on new comments "
            + "(default: %(default)s, use 0 to set no maximum line length)",
        )
        parser.add_argument(
            "--truncate",
            action="store_true",
            help="Truncate error messages to maximum line length.",
        )
        parser.add_argument(
            "--unsafe",
            action="store_true",
            help="Don't check syntax when applying fixmes.",
        )
        parser.add_argument(
            "--force-format-unsuppressed", action="store_true", help=argparse.SUPPRESS
        )
        parser.add_argument(
            "--lint",
            action="store_true",
            help="Run lint to ensure added fixmes comply with black formatting. \
            Doubles the runtime of pyre-ugprade.",
        )
        parser.add_argument("--no-commit", action="store_true", help=argparse.SUPPRESS)
        parser.add_argument(
            "--do-not-run-buck-clean",
            action="store_false",
            dest="should_clean",
            default=True,
            help=argparse.SUPPRESS,
        )

    def _apply_suppressions(self, errors: Errors) -> None:
        try:
            errors.suppress(
                self._comment,
                self._max_line_length,
                self._truncate,
                self._unsafe,
            )
        except PartialErrorSuppression as partial_error_suppression:
            if not self._force_format_unsuppressed:
                raise partial_error_suppression
            self._repository.force_format(partial_error_suppression.unsuppressed_paths)
            errors.suppress(
                self._comment,
                self._max_line_length,
                self._truncate,
                self._unsafe,
            )

    def _get_and_suppress_errors(
        self,
        configuration: Configuration,
        error_source: ErrorSource = ErrorSource.GENERATE,
        upgrade_version: bool = False,
        only_fix_error_code: Optional[int] = None,
        fixme_threshold: Optional[int] = None,
        fixme_threshold_fallback_mode: LocalMode = LocalMode.IGNORE,
    ) -> None:
        LOG.info("Processing %s", configuration.get_directory())
        if not configuration.is_local:
            return
        if upgrade_version:
            if configuration.version:
                configuration.remove_version()
                configuration.write()
            else:
                return
        errors = (
            Errors.from_stdin(only_fix_error_code)
            if error_source == ErrorSource.STDIN and not upgrade_version
            else configuration.get_errors(
                only_fix_error_code, should_clean=self._should_clean
            )
        )
        if len(errors) == 0:
            return

        if fixme_threshold is None:
            self._apply_suppressions(errors)
        else:
            for path, path_errors in errors.paths_to_errors.items():
                path_errors = list(path_errors)
                if len(path_errors) > fixme_threshold:
                    LOG.info(
                        "%d errors found in `%s`. Adding file-level ignore.",
                        len(path_errors),
                        path,
                    )
                    add_local_mode(path, fixme_threshold_fallback_mode)
                else:
                    self._apply_suppressions(Errors(path_errors))

        # Lint and re-run pyre once to resolve most formatting issues
        if self._lint and self._repository.format():
            errors = configuration.get_errors(only_fix_error_code, should_clean=False)
            self._apply_suppressions(errors)
